package com.softslate.commerce.businessobjects.order;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.BaseBusinessProcessor;
import com.softslate.commerce.businessobjects.product.Discount;
import com.softslate.commerce.businessobjects.product.DiscountRange;
import com.softslate.commerce.businessobjects.product.Product;
import com.softslate.commerce.businessobjects.product.ProductAttribute;
import com.softslate.commerce.businessobjects.product.SKU;
import com.softslate.commerce.businessobjects.product.SKUDiscount;
import com.softslate.commerce.daos.order.OrderDAO;
import com.softslate.commerce.daos.order.OrderGatewayDAO;
import com.softslate.commerce.daos.product.DiscountDAO;

/**
 * Processes business logic having to do with discounts in a user's cart.
 * <p>
 * <code>BasicCartDiscountProcessor</code> is the default implementation of the
 * <code>CartDiscountProcessor</code> interface for the application.
 * 
 * @author David Tobey
 * 
 * @see com.softslate.commerce.businessobjects.order.InventoryProcessor
 */
public class BasicCartDiscountProcessor extends BaseBusinessProcessor implements
		CartDiscountProcessor {

	static Log log = LogFactory.getLog(BasicCartDiscountProcessor.class);

	boolean justProcessGiftCertificates = false;

	public Map processAddItems(Collection newOrderItems) throws Exception {
		return processDiscounts(newOrderItems);
	}

	public Map processCouponCode(Map parameters) throws Exception {
		return processDiscounts(null);
	}

	public Map processOnCheckout(Map parameters) throws Exception {
		return processDiscounts(null);
	}

	public Map processAfterTaxes(Map parameters) throws Exception {
		justProcessGiftCertificates = true;
		return processDiscounts(null);
	}

	public Map processCustomerLogin(Map parameters) throws Exception {
		return processDiscounts(null);
	}

	public Map processCustomerLogout(Map parameters) throws Exception {
		return processDiscounts(null);
	}

	public Map processOrderComplete(Map parameters) throws Exception {

		// Too late to reprocess the discounts since the user has already
		// submitted payment information and it's been processed. Just update
		// the number of times used field.
		return processTimesUsed();
	}

	/**
	 * Processes discount for the current user and his or her cart. This method
	 * calls the following methods in turn:
	 * <ol>
	 * <li>{@link #processRemoveDiscounts()}</li>
	 * <li>{@link #processSkuDiscounts(Collection)}</li>
	 * <li>{@link #processGlobalDiscounts()}</li>
	 * </ol>
	 * Note all the discounts already in the user's cart are removed, then added
	 * back in.
	 * 
	 * @param newOrderItems
	 *            The <code>Collection</code> of <code>OrderItem</code>s in the
	 *            user's cart.
	 * @return A <code>Map</code> containing results of the processing. This
	 *         implementation returns null.
	 * @throws Exception
	 */
	public Map processDiscounts(Collection newOrderItems) throws Exception {
		if (log.isDebugEnabled())
			log.debug("Processing discounts.");
		String applyDiscounts = (String) getSettings().getValue(
				"applyDiscounts");
		if (getUser().getOrder() != null
				&& getUser().getOrder().firstDelivery() != null
				&& applyDiscounts != null && applyDiscounts.equals("1")) {

			// If the order is already completed (ie, this processing was
			// triggered from the admin area, we must first decrement the "times
			// used" fields for each of the discounts
			if (getUser().getOrder().getCompleted() != null) {
				processTimesUsed(-1);
			}

			// Remove all existing order discounts
			processRemoveDiscounts();

			if (!justProcessGiftCertificates) {

			// Do the sku discounts first
			// Use the existing order items in the user's cart (the new items
			// should already have been added)
			Collection orderItems = getUser().getOrder().firstDelivery()
					.getOrderItemCollection();
			processSkuDiscounts(orderItems);

			// Then the global discounts
			processGlobalDiscounts();

			// Then the shipping discounts
			processShippingDiscounts();
			}

			// Then the gift certificates
			processGiftCertificates();

			// Update the order in case it is detached from the Hibernate
			// session. Commit the transaction in case discounts are processed
			// multiple times in the same request.
			OrderDAO orderDAO = (OrderDAO) getDaoFactory().createDAO("orderDAOImplementer");
			orderDAO.setOrder(getUser().getOrder());
			orderDAO.updateOrder(true);

			// If the order is already completed (ie, this processing was
			// triggered from the admin area, we must immediately increment the
			// times used
			if (getUser().getOrder().getCompleted() != null) {
				processTimesUsed();
			}

			// Update the order in case it is detached from the Hibernate
			// session. Commit the transaction in case discounts are processed
			// multiple times in the same request.
			orderDAO = (OrderDAO) getDaoFactory().createDAO(
					"orderDAOImplementer");
			orderDAO.setOrder(getUser().getOrder());
			orderDAO.updateOrder(true);

		}
		return null;
	}

	public Map processRemoveDiscounts() throws Exception {
		if (log.isDebugEnabled())
			log.debug("Removing all existing discounts for the order.");

		// Add up the totals of all the discounts in the cart
		Collection orderDiscounts = getUser().getOrder().getOrderDiscounts();
		if (orderDiscounts == null) {
			return null;
		}
		double discountSubTotal = 0;
		double discountTotal = 0;
		boolean removeIt = false;
		Iterator i = orderDiscounts.iterator();
		while (i.hasNext()) {
			OrderDiscount od = (OrderDiscount) i.next();
			if (od.getItemsAppliesType().equals("[giftCertificate]")) {
				discountTotal = discountTotal + od.getTotal().doubleValue();
				removeIt = true;
			} else if (od.getItemsAppliesType().equals("[shipping]")) {

				// Note: shipping discounts are applied during shipping
				// processing - nothing to do here.
			} else if (!justProcessGiftCertificates) {
				discountSubTotal = discountSubTotal + od.getTotal().doubleValue();
				removeIt = true;
			}
			if (removeIt) {
				if (od.getOrderItem() != null && od.getOrderItem().getOrderDiscounts() != null) {
				od.getOrderItem().getOrderDiscounts().remove(od);
			}
			i.remove();
		}
		}

		// Increment totals of the Order and OrderDelivery.
		if (log.isDebugEnabled())
			log.debug("Adding discounts back into cart totals.");
		if (getUser().getOrder().getTotal() != null) {
			getUser().getOrder().setTotal(
					new Double(getUser().getOrder().getTotal().doubleValue()
							+ discountTotal));
		}
		if (getUser().getOrder().firstDelivery().getTotal() != null) {
			getUser()
					.getOrder()
					.firstDelivery()
					.setTotal(
							new Double(getUser().getOrder().firstDelivery()
									.getTotal().doubleValue()
									+ discountTotal));
		}
		if (log.isDebugEnabled())
			log.debug("Adding discounts back into cart subtotals.");
		getUser().getOrder().setSubtotal(
				new Double(getUser().getOrder().getSubtotal().doubleValue()
						+ discountSubTotal));
		getUser()
				.getOrder()
				.firstDelivery()
				.setSubtotal(
						new Double(getUser().getOrder().firstDelivery()
								.getSubtotal().doubleValue()
								+ discountSubTotal));

		// Add formatting to the memory objects
		getUser()
				.getOrder()
				.firstDelivery()
				.addFormatting(getUser().getLocale(),
						(String) getSettings().getValue("currencyCode"));
		getUser().getOrder().addFormatting(getUser().getLocale(),
				(String) getSettings().getValue("currencyCode"));

		return null;
	}

	/**
	 * Processes discounts associated with specific SKUs in the current user's
	 * cart.
	 * 
	 * @param orderItems
	 *            The <code>Collection</code> of <code>OrderItem</code>s in the
	 *            user's cart.
	 * @return A <code>Map</code> containing results of the processing. This
	 *         implementation returns null.
	 * @throws Exception
	 */
	public Map processSkuDiscounts(Collection orderItems) throws Exception {
		if (log.isDebugEnabled())
			log.debug("Processing SKU discounts.");
		if (orderItems == null || orderItems.isEmpty()) {
			return null;
		}

		// Initialize some variables
		Double previousOrderTotal = null;
		Collection skuDiscountsToApply = new ArrayList();
		Collection productIDs = getProductIDs(orderItems);
		Collection attributeIDs = getAttributeIDs(orderItems);
		Collection matchingSkus = loadMatchingSkus(orderItems, productIDs,
				attributeIDs);

		// If no skus match, we're done
		if (matchingSkus == null || matchingSkus.isEmpty()) {
			return null;
		}

		// Collect all the discounts to be applied first, in
		// skuDiscountsToApply, while associating their order items with them.
		Iterator it = orderItems.iterator();
		while (it.hasNext()) {
			OrderItem oi = (OrderItem) it.next();
			Collection skuIDs = getSkuIDs(oi, matchingSkus);
			if (skuIDs == null || skuIDs.isEmpty()) {
				continue;
			}
			Collection discounts = getDiscountsBySkuIDs(skuIDs);
			if (discounts == null || discounts.isEmpty()) {
				continue;
			}
			// We found a discount
			Iterator id = discounts.iterator();
			while (id.hasNext()) {
				Discount discount = (Discount) id.next();
				if (isDiscountOk(discount, previousOrderTotal)) {
					if (log.isDebugEnabled())
						log.debug("Applicable discount: " + discount.getCode()
								+ " order item: " + oi.getCode());
					addSkuDiscount(skuDiscountsToApply, discount, oi);
				}
			}

		}
		// We should now have a collection of discounts to be applied.
		// Each discount to be applied should hold a collection of order items
		// that also match. It should also have updated accumulated totals.
		// Now have it do all the internal calculations so each
		// Discount will have a map of discounting information
		if (log.isDebugEnabled())
			log.debug("Calculating SKU discounts");
		Iterator i = skuDiscountsToApply.iterator();
		while (i.hasNext()) {
			Discount ad = (Discount) i.next();
			Map results = calculateDiscount(ad, ad.getTotalAmount(),
					ad.getTotalQuantity());
			ad.setDiscountInformation(results);
		}

		// Loop through the order items.
		if (log.isDebugEnabled())
			log.debug("Creating order discounts");
		Iterator oii1 = orderItems.iterator();
		while (oii1.hasNext()) {
			OrderItem oi1 = (OrderItem) oii1.next();
			if (log.isDebugEnabled())
				log.debug("Processing order item: " + oi1.getCode() + " id: "
						+ oi1.getOrderItemID());

			// Keep track of the total amount being discounted from this order
			// item so we can make it doesn't discount more than the item is
			// worth.
			double totalToDiscount = 0;

			// We loop through the discounts and loop through
			// the oi's held by the discounts and apply appropriately
			Iterator adi = skuDiscountsToApply.iterator();
			while (adi.hasNext()) {
				Discount ad = (Discount) adi.next();
				if (log.isDebugEnabled())
					log.debug("Processing discount: " + ad.getCode());
				Iterator oii2 = ad.getOrderItems().iterator();
				while (oii2.hasNext()) {
					OrderItem oi2 = (OrderItem) oii2.next();
					if (log.isDebugEnabled())
						log.debug("Looping through order item: "
								+ oi2.getCode());
					if (oi2.equals(oi1)) {
						if (log.isDebugEnabled())
							log.debug("Adding order discount for discount: "
									+ ad.getCode() + " and order item: "
									+ oi1.getCode());

						// We have a matching order item
						Map results = ad.getDiscountInformation();
						Double discountTotal = (Double) results
								.get("discountTotal");
						Double floor = (Double) results.get("floor");
						Double ceiling = (Double) results.get("ceiling");
						Double reward = (Double) results.get("reward");
						String rewardType = (String) results.get("rewardType");
						if (results != null && discountTotal != null) {
							if (log.isDebugEnabled())
								log.debug("Adding discount: " + ad.getCode());

							OrderDiscount od =  getBusinessObjectFactory()
									.createObject(OrderDiscount.class);
							BeanUtils.copyProperties(od, ad);

							// If this is a percentage, we try to calculate
							// against the order item's total.
							// If it is an amount type, we
							// can do it by just getting a discount per item and
							// subtracting out accordingly.
							if (rewardType.equals("[percentage]")) {
								discountTotal = new Double(
										(reward.doubleValue() * oi1.getTotal()
												.doubleValue()) / 100);
								discountTotal = utils().roundForMoney(
										discountTotal);
							} else {
								// We have to calculate the total per item
								Double discountPerItem = new Double(
										discountTotal.doubleValue()
												/ ad.getTotalQuantity()
														.intValue());
								discountTotal = new Double(
										discountPerItem.doubleValue()
												* oi1.getQuantity());
								discountTotal = utils().roundForMoney(
										discountTotal);
							}

							// Ensure the total amount being discounted for this
							// item doesn't exceed the item's total.
							if (totalToDiscount + discountTotal.doubleValue() > oi1
									.getTotal().doubleValue()) {
								discountTotal = new Double(oi1.getTotal()
										.doubleValue() - totalToDiscount);
							}
							totalToDiscount = totalToDiscount
									+ discountTotal.doubleValue();

							// Populate the OrderDiscount
							od.setTotal(discountTotal);
							od.setFloor(floor);
							od.setCeiling(ceiling);
							od.setReward(reward);
							od.setRewardType(rewardType);
							od.setOrder(getUser().getOrder());
							od.setOrderItem(oi1);

							od.setName(od.getName());
							if (getUser().getOrder().getOrderDiscounts() == null) {
								getUser().getOrder().setOrderDiscounts(
										new LinkedHashSet());
							}
							getUser().getOrder().getOrderDiscounts().add(od);
							if (oi1.getOrderDiscounts() == null) {
								oi1.setOrderDiscounts(new LinkedHashSet());
							}
							oi1.getOrderDiscounts().add(od);

							// Update the order
							updateOrderAndDeliveryTotals(discountTotal);

						} // Apply Discount
					} // End if OI's match
				} // End looping over advanced discount order items
			} // End looping over advanced discounts
		} // End looping over order items
		return null;
	}

	/**
	 * Processes global discounts, which are not associated with specific SKUs.
	 * 
	 * @return A <code>Map</code> containing results of the processing. This
	 *         implementation returns null.
	 * @throws Exception
	 */
	public Map processGlobalDiscounts() throws Exception {
		if (log.isDebugEnabled())
			log.debug("Processing global discounts");
		if (getUser().getOrder().getSubtotal() == null
				|| getUser().getOrder().getQuantity() == null) {
			return null;
		}

		// Initialize some variables
		Double previousOrderTotal = null;
		Collection discounts = getGlobalDiscounts();

		// Loop through the global discounts
		Iterator j = discounts.iterator();
		while (j.hasNext()) {
			Discount discount = (Discount) j.next();

			// Should this discount be applied?
			if (log.isDebugEnabled())
				log.debug("Checking if disc applies: " + discount.getCode()
						+ "; " + previousOrderTotal);
			if (!isDiscountOk(discount, previousOrderTotal)) {
				continue;
			}

			// Calculate the discount
			if (log.isDebugEnabled())
				log.debug("Calculating discount total: " + discount.getCode()
						+ "; " + getUser().getOrder().getSubtotal() + "; "
						+ getUser().getOrder().getQuantity());
			Map results = calculateDiscount(discount, getUser().getOrder()
					.getSubtotal(), getUser().getOrder().getQuantity());
			Double discountTotal = (Double) results.get("discountTotal");
			Double floor = (Double) results.get("floor");
			Double ceiling = (Double) results.get("ceiling");
			Double reward = (Double) results.get("reward");
			String rewardType = (String) results.get("rewardType");
			if (discountTotal == null) {
				continue;
			}

			// Create the order discount and add it to the order
			if (log.isDebugEnabled())
				log.debug("Adding discount: " + discount.getCode());
			OrderDiscount od =  getBusinessObjectFactory()
					.createObject(OrderDiscount.class);
			BeanUtils.copyProperties(od, discount);
			od.setTotal(discountTotal);
			od.setFloor(floor);
			od.setCeiling(ceiling);
			od.setReward(reward);
			od.setRewardType(rewardType);
			od.setOrder(getUser().getOrder());
			od.setOrderItemID(null);
			od.setOrderItem(null);
			if (getUser().getOrder().getOrderDiscounts() == null) {
				getUser().getOrder().setOrderDiscounts(new LinkedHashSet());
			}
			getUser().getOrder().getOrderDiscounts().add(od);

			// Subtract the discount total from the order's and first
			// delivery's subtotals
			updateOrderAndDeliveryTotals(discountTotal);

		}
		return null;
	}

	/**
	 * Queries for and inserts the shipping discounts applicable for the
	 * user/order. The discount is not actually applied - a zero-total discount
	 * is inserted. The discount is actually applied during shipping processing,
	 * ie, processShippingDiscountsOnShippingOptions()
	 * 
	 * @return A Map, null in this implementation.
	 * @throws Exception
	 */
	public Map processShippingDiscounts() throws Exception {
		if (log.isDebugEnabled())
			log.debug("Processing shipping discounts.");
		if (getUser().getOrder().getSubtotal() == null
				|| getUser().getOrder().getQuantity() == null) {
			return null;
		}

		// Initialize some variables
		Double previousOrderTotal = null;
		Collection discounts = loadShippingDiscounts();

		// Loop through the shipping discounts
		Iterator j = discounts.iterator();
		while (j.hasNext()) {
			Discount discount = (Discount) j.next();

			// Has this discount already been added?
			if (hasDiscountBeenAddedToOrder(getUser().getOrder(), discount)) {
				if (log.isInfoEnabled())
					log.info("Shipping disc has already been added: " + discount.getCode());
				continue;
			}

			// Should this discount be applied?
			if (log.isInfoEnabled())
				log.info("Checking if shipping disc applies: " + discount.getCode()
						+ "; " + previousOrderTotal);
			if (!isDiscountOk(discount, previousOrderTotal)) {
				continue;
			}

			// The total of the discount should be 0 or null. 0 if the discount
			// applies, and null if it does not. 0 here because the amount is
			// subtracted during shipping processing directly from
			// Order.shipping.
			Double totalToDiscount = new Double(0);
			if (log.isDebugEnabled())
				log.debug("Total possible to discount: " + totalToDiscount);

			// Calculate the discount
			if (log.isDebugEnabled())
				log.debug("Calculating discount total: " + discount.getCode()
						+ "; " + totalToDiscount + "; "
						+ getUser().getOrder().getQuantity());
			Map results = calculateDiscount(discount, totalToDiscount,
					getUser().getOrder().getQuantity());
			Double discountTotal = (Double) results.get("discountTotal");
			Double floor = (Double) results.get("floor");
			Double ceiling = (Double) results.get("ceiling");
			Double reward = (Double) results.get("reward");
			String rewardType = (String) results.get("rewardType");
			if (discountTotal == null) {
				continue;
			}

			// Create the order discount and add it to the order
			if (log.isDebugEnabled())
				log.debug("Adding discount: " + discount.getCode());
			OrderDiscount od =  getBusinessObjectFactory()
					.createObject(OrderDiscount.class);
			BeanUtils.copyProperties(od, discount);
			od.setTotal(discountTotal);
			od.setFloor(floor);
			od.setCeiling(ceiling);
			od.setReward(reward);
			od.setRewardType(rewardType);
			od.setOrder(getUser().getOrder());
			if (getUser().getOrder().getOrderDiscounts() == null) {
				getUser().getOrder().setOrderDiscounts(new LinkedHashSet());
			}
			getUser().getOrder().getOrderDiscounts().add(od);

			// Note the discount is actually applied to the order's shipping
			// amount during shipping processing
		}
		return null;
	}

	protected boolean hasDiscountBeenAddedToOrder(Order order, Discount discount) {
		boolean ret = false;
		if (order != null && order.getOrderDiscounts() != null) {
			for (OrderDiscount od : (Collection<OrderDiscount>) order.getOrderDiscounts()) {
				if (od.getCode().equals(discount.getCode())) {
					ret = true;
					break;
				}
			}
		}
		return ret;
	}

	public Collection processShippingDiscountsOnShippingOptions(
			Collection shippingOptions) throws Exception {
		if (shippingOptions == null || shippingOptions.isEmpty()
				|| getUser().getOrder() == null) {
			return null;
		}
		if (log.isInfoEnabled())
			log.info("Applying shipping discounts to shipping options.");

		// Initialize some variables
		Double previousOrderTotal = null;
		Collection discounts = loadShippingDiscounts();
		String currencyCode = (String) getSettings().getValue("currencyCode");
		Double savedShipping = getUser().getOrder().getShipping();
		String savedShippingMethod = getUser().getOrder().getShippingMethod();
		try {
			Iterator i = shippingOptions.iterator();
			while (i.hasNext()) {
				Map option = (Map) i.next();
				Double optionPrice = (Double) option.get("priceDouble");
				String optionCode = (String) option.get("code");
				String optionName = (String) option.get("name");
				Collection optionDiscounts = (Collection) option
						.get("discounts");
				getUser().getOrder().setShipping(optionPrice);
				getUser().getOrder().setShippingMethod(optionCode);
				if (optionName != null
						&& getSettings().getValue(
								"storeShippingMethodNameWithOrder") != null
						&& getSettings().getValue(
								"storeShippingMethodNameWithOrder").equals("1")) {
					getUser().getOrder().setShippingMethod(optionName);
				}
				if (log.isDebugEnabled())
					log.debug("Applying shipping discounts to option: "
							+ optionName);

				// Loop through the shipping discounts
				Iterator j = discounts.iterator();
				while (j.hasNext()) {
					Discount discount = (Discount) j.next();

					// Should this discount be applied?
					if (log.isDebugEnabled())
						log.debug("Checking if disc applies: "
								+ discount.getCode() + "; "
								+ previousOrderTotal);
					if (!isDiscountOk(discount, previousOrderTotal)) {
						continue;
					}

					// Decrement from the shipping amount
					Double totalToDiscount = getUser().getOrder().getShipping();
					if (log.isDebugEnabled())
						log.debug("Total possible to discount: "
								+ totalToDiscount);

					// Calculate the discount
					if (log.isDebugEnabled())
						log.debug("Calculating discount total: "
								+ discount.getCode() + "; " + totalToDiscount
								+ "; " + getUser().getOrder().getQuantity());
					Map results = calculateDiscount(discount, totalToDiscount,
							getUser().getOrder().getQuantity());
					Double discountTotal = (Double) results
							.get("discountTotal");
					if (discountTotal == null
							|| discountTotal.equals(Double.valueOf(0))) {
						continue;
					}

					// Apply the discount to the shipping option
					optionPrice = new Double(optionPrice.doubleValue()
							- discountTotal.doubleValue());
					option.put("priceDouble", optionPrice);
					option.put(
							"price",
							formatPrice(optionPrice.doubleValue(),
									currencyCode, getUser().getLocale()));
					if (optionDiscounts == null) {
						optionDiscounts = new ArrayList();
					}
					optionDiscounts.add(BeanUtils.describe(discount));
					option.put("discounts", optionDiscounts);
				}
			}
		} finally {
			getUser().getOrder().setShipping(savedShipping);
			getUser().getOrder().setShippingMethod(savedShippingMethod);
		}
		return shippingOptions;
	}

	/**
	 * Queries for and applies the gift certs applicable for the user/order.
	 * 
	 * @return A Map, null in this implementation.
	 * @throws Exception
	 */
	public Map processGiftCertificates() throws Exception {
		if (log.isDebugEnabled())
			log.debug("Processing gift certificates. "
					+ "(The reward must be [amount] not %.)");
		if (getUser().getOrder().getSubtotal() == null
				|| getUser().getOrder().getQuantity() == null) {
			return null;
		}

		// Initialize some variables
		Double previousOrderTotal = null;
		Collection discounts = loadGiftCertificates();

		// Loop through the gift certificates
		Iterator j = discounts.iterator();
		while (j.hasNext()) {
			Discount discount = (Discount) j.next();

			// Should this discount by applied?
			if (log.isDebugEnabled())
				log.debug("Checking if disc applies: " + discount.getCode()
						+ "; " + previousOrderTotal);
			if (!isDiscountOk(discount, previousOrderTotal)) {
				continue;
			}

			// Check that the reward is not all used up
			if (discount.getReward() != null
					&& discount.getReward().doubleValue() <= 0) {
				if (log.isInfoEnabled())
					log.info("Gift certificate "
							+ discount.getCouponCode()
							+ " cannot be used because its balance is 0 or less");
				continue;
			}

			// Decrement from the order total if defined, otherwise the order's
			// subtotal
			Double totalToDiscount = getUser().getOrder().getSubtotal();
			if (getUser().getOrder().getTotal() != null) {
				if (log.isDebugEnabled())
					log.debug("Order total defined.");
				totalToDiscount = getUser().getOrder().getTotal();
			}
			if (log.isDebugEnabled())
				log.debug("Total possible to discount: " + totalToDiscount);

			// Calculate the discount
			if (log.isDebugEnabled())
				log.debug("Calculating discount total: " + discount.getCode()
						+ "; " + totalToDiscount + "; "
						+ getUser().getOrder().getQuantity());
			Map results = calculateDiscount(discount, totalToDiscount,
					getUser().getOrder().getQuantity());
			Double discountTotal = (Double) results.get("discountTotal");
			Double floor = (Double) results.get("floor");
			Double ceiling = (Double) results.get("ceiling");
			Double reward = (Double) results.get("reward");
			String rewardType = (String) results.get("rewardType");
			if (discountTotal == null) {
				continue;
			}

			// Create the order discount and add it to the order
			if (log.isDebugEnabled())
				log.debug("Adding discount: " + discount.getCode());
			OrderDiscount od =  getBusinessObjectFactory()
					.createObject(OrderDiscount.class);
			BeanUtils.copyProperties(od, discount);
			od.setTotal(discountTotal);
			od.setFloor(floor);
			od.setCeiling(ceiling);
			od.setReward(reward);
			od.setRewardType(rewardType);
			od.setOrder(getUser().getOrder());
			od.setOrderItemID(null);
			od.setOrderItem(null);
			if (getUser().getOrder().getOrderDiscounts() == null) {
				getUser().getOrder().setOrderDiscounts(new LinkedHashSet());
			}
			getUser().getOrder().getOrderDiscounts().add(od);

			// Subtract the discount total from the order's and first
			// delivery's totals - not subtotals
			updateGiftCertOrderAndDeliveryTotals(discountTotal);

		}
		return null;
	}

	public Map processTimesUsed() throws Exception {
		return processTimesUsed(1);
	}

	public Map processTimesUsed(int addOrSubtract) throws Exception {
		if (log.isDebugEnabled())
			log.debug("Updating the times used field for discounts.");
		Collection orderDiscounts = getUser().getOrder().getOrderDiscounts();
		if (orderDiscounts == null) {
			return null;
		}
		DiscountDAO dao = (DiscountDAO) getDaoFactory().createDAO(
				"discountDAOImplementer");

		// Loop through each of the discounts in the order
		Iterator i = orderDiscounts.iterator();
		while (i.hasNext()) {
			OrderDiscount od = (OrderDiscount) i.next();
			if (od.getDiscountID() != null) {

				// Load the corresponding Discount object
				Discount d =  getBusinessObjectFactory()
						.createObject(Discount.class);
				d.setDiscountID(od.getDiscountID().intValue());
				dao.setDiscount(d);
				dao.loadDiscountFromID();
				if (dao.getDiscount() != null) {

					// If it's a gift certificate, decrement (or add to) the
					// reward amount and deactivate if it's used up (or activate
					// if it now has a balance).
					if (dao.getDiscount().getItemsAppliesType() != null
							&& dao.getDiscount().getItemsAppliesType()
									.equals("[giftCertificate]")
							&& dao.getDiscount().getRewardType() != null
							&& dao.getDiscount().getRewardType()
									.equals("[amount]")
							&& dao.getDiscount().getReward() != null) {

						// The amount to decrement/increment
						double amount = addOrSubtract
								* od.getTotal().doubleValue();

						// Decrement the reward amount here
						dao.getDiscount().setReward(
								new Double(dao.getDiscount().getReward()
										.doubleValue()
										- amount));
					}

					// For all discounts, add 1 to the times used value
					if (dao.getDiscount().getTimesUsed() == null) {
						dao.getDiscount().setTimesUsed(new Integer(0));
					}
					dao.getDiscount().setTimesUsed(
							new Integer(dao.getDiscount().getTimesUsed()
									.intValue()
									+ addOrSubtract));
					dao.updateDiscount();
				}
			}
		}
		return null;
	}

	public Collection getSkuIDs(OrderItem oi, Collection matchingSkus) {
		Collection skuIDs = new ArrayList();
		Iterator j = matchingSkus.iterator();
		while (j.hasNext()) {
			Map sku = (Map) j.next();
			if (oi.equals((OrderItem) sku.get("orderItem"))) {
				skuIDs.add(new Integer((String) sku.get("skuID")));
			}
		}
		return skuIDs;
	}

	/**
	 * Returns a <code>Collection</code> of
	 * {@link com.softslate.commerce.businessobjects.product.Discount} objects
	 * associated with the given <code>Collection</code> of
	 * {@link com.softslate.commerce.businessobjects.product.SKU} ids. Used
	 * during SKU discount processing.
	 * 
	 * @param skuIDs
	 *            A <code>Collection</code> of
	 *            {@link com.softslate.commerce.businessobjects.product.SKU}
	 *            ids.
	 * @return A <code>Collection</code> of
	 *         {@link com.softslate.commerce.businessobjects.product.Discount}
	 *         objects
	 * @throws Exception
	 */
	public Collection getDiscountsBySkuIDs(Collection skuIDs) throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory()
				.createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.getDiscountsBySkuIDs(skuIDs);
	}

	/**
	 * 
	 * Returns a <code>Collection</code> of global
	 * {@link com.softslate.commerce.businessobjects.product.Discount} objects
	 * that are not associated with a particular <code>SKU</code>. Used during
	 * global discount processing.
	 * 
	 * @return A <code>Collection</code> of
	 *         {@link com.softslate.commerce.businessobjects.product.Discount}
	 *         objects
	 * @throws Exception
	 */
	public Collection getGlobalDiscounts() throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory()
				.createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.getGlobalDiscounts(getUser()
				.getEnteredCouponCodes());
	}

	/**
	 * 
	 * Returns a <code>Collection</code> of shipping discounts -
	 * {@link com.softslate.commerce.businessobjects.product.Discount} objects
	 * matching the list of coupon codes that the current user has entered. Used
	 * during discount processing.
	 * 
	 * @return A <code>Collection</code> of
	 *         {@link com.softslate.commerce.businessobjects.product.Discount}
	 *         objects
	 * @throws Exception
	 */
	public Collection loadShippingDiscounts() throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory()
				.createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadShippingDiscounts(getUser()
				.getEnteredCouponCodes());
	}

	/**
	 * 
	 * Returns a <code>Collection</code> of gift certificate
	 * {@link com.softslate.commerce.businessobjects.product.Discount} objects
	 * matching the list of coupon/gift certificate codes that the current user
	 * has entered. Used during discount processing.
	 * 
	 * @return A <code>Collection</code> of
	 *         {@link com.softslate.commerce.businessobjects.product.Discount}
	 *         objects
	 * @throws Exception
	 */
	public Collection loadGiftCertificates() throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory()
				.createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadGiftCertificates(getUser()
				.getEnteredCouponCodes());
	}

	/**
	 * Checks that a given discount is applicable. Used by discount processing
	 * to help determine which discounts to apply, and in the presentation to
	 * determine which discounts to display to the user.
	 * <p>
	 * This method calls the following methods in turn:
	 * <ol>
	 * <li>{@link #isStartedAndNotExpired(Discount)}</li>
	 * <li>{@link #isCouponOk(Discount)}</li>
	 * <li>{@link #isUserQualified(Discount, Double)}</li>
	 * <li>{@link #isCartItemsOk(Discount)}</li>
	 * <li>{@link #isTimesUsedOk(Discount)}</li>
	 * </ol>
	 * 
	 * @param discount
	 *            The discount whose applicability is being checked.
	 * @param previousOrderTotal
	 *            The previous order total of the current user, or null. Some
	 *            discount are only applicable if the user has previously placed
	 *            orders totaling a certain amount.
	 * @return true, the discount is applicable, otherwise, false.
	 * @throws Exception
	 */
	public boolean isDiscountOk(Discount discount, Double previousOrderTotal)
			throws Exception {
		if (!discount.getIsActive()) {
			return false;
		}
		if (!isStartedAndNotExpired(discount)) {
			return false;
		}
		if (!isCouponOk(discount)) {
			return false;
		}
		if (!isUserQualified(discount, previousOrderTotal)) {
			return false;
		}
		if (!isCartItemsOk(discount)) {
			return false;
		}
		if (!isShippingMethodOk(discount)) {
			return false;
		}
		if (!isTimesUsedOk(discount)) {
			return false;
		}
		return true;
	}

	/**
	 * Checks if a given discount would be applicable, if the given product code
	 * were in the user's cart. Used by the presentation layer to communicate
	 * the discounts that would be available if the item were added to the cart.
	 * <p>
	 * This method calls the following methods in turn:
	 * <ol>
	 * <li>{@link #isStartedAndNotExpired(Discount)}</li>
	 * <li>{@link #isCouponOk(Discount)}</li>
	 * <li>{@link #isUserQualified(Discount, Double)}</li>
	 * <li>{@link #isTimesUsedOk(Discount)}</li>
	 * </ol>
	 * 
	 * @param discount
	 *            The discount whose applicability is being checked.
	 * @param previousOrderTotal
	 *            The previous order total of the current user, or null. Some
	 *            discount are only applicable if the user has previously placed
	 *            orders totaling a certain amount.
	 * @return true, the discount is applicable, otherwise, false.
	 * @throws Exception
	 */
	public boolean isDiscountOkIfItemInCart(Discount discount,
			Double previousOrderTotal, String productCode) throws Exception {
		if (!discount.getIsActive()) {
			return false;
		}
		if (!isStartedAndNotExpired(discount)) {
			return false;
		}
		if (!isCouponOk(discount)) {
			return false;
		}
		if (!isUserQualified(discount, previousOrderTotal)) {
			return false;
		}
		if (!isItemWouldAllowDiscount(discount, productCode)) {
			return false;
		}
		if (!isTimesUsedOk(discount)) {
			return false;
		}
		return true;
	}

	/**
	 * Checks that the current time is not before a given
	 * {@link com.softslate.commerce.businessobjects.product.Discount}'s start
	 * date or after its expiration date.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @return true, if the current time is not before the <code>Discount</code>
	 *         's start date or after its expiration date.
	 * @throws Exception
	 */
	public boolean isStartedAndNotExpired(Discount discount) throws Exception {
		Date now = new Date();

		// If the order is already complete (ie, this processing being triggered
		// from the admin area), use the order's completed date instead of now.
		if (getUser().getOrder() != null
				&& getUser().getOrder().getCompleted() != null) {
			now = parseDateTime(getUser().getOrder().getCompleted());
		}
		boolean isStartOk = false;
		if (discount.getStartDate() == null
				|| discount.getStartDate().equals("")) {
			isStartOk = true;
		} else {
			Date start = parseDateTime(discount.getStartDate());
			if (start.before(now)) {
				isStartOk = true;
			}
		}
		if (!isStartOk) {
			if (log.isDebugEnabled())
				log.debug("Discount not ok - not yet the start date.");
			return false;
		}
		boolean isExpirationOk = false;
		if (discount.getExpirationDate() == null
				|| discount.getExpirationDate().equals("")) {
			isExpirationOk = true;
		} else {
			Date expire = parseDateTime(discount.getExpirationDate());
			if (expire.after(now)) {
				isExpirationOk = true;
			}
		}
		if (!isExpirationOk) {
			if (log.isDebugEnabled())
				log.debug("Discount not ok - after the expiration date.");
			return false;
		}
		return true;
	}

	/**
	 * Checks that a given
	 * {@link com.softslate.commerce.businessobjects.product.Discount} has not
	 * been used more than the maximum number of times it's allowed to be used.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @return true, if the <code>Discount</code> has not been used more than
	 *         the maximum number of times it's allowed to be used.
	 * @throws Exception
	 */
	public boolean isTimesUsedOk(Discount discount) throws Exception {
		if (!discount.getIsTimesPerCustomer()) {
			if (!(discount.getMaxTimesUsed() == null
					|| discount.getTimesUsed() == null || discount
					.getMaxTimesUsed().intValue() == 0)
					&& discount.getMaxTimesUsed().intValue() <= discount
							.getTimesUsed().intValue()) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - used too many times.");
				return false;
			}
		} else {
			if (getUser().getCustomer() == null) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - times used is by customer but customer not logged in.");
				return false;
			}
			Integer timesUsed = getTimesUsedByCustomer(discount);
			if (!(discount.getMaxTimesUsed() == null || timesUsed == null || discount
					.getMaxTimesUsed().intValue() == 0)
					&& discount.getMaxTimesUsed().intValue() <= timesUsed
							.intValue()) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - used too many times by cust.");
				return false;
			}
		}
		return true;
	}

	/**
	 * Checks that a given
	 * {@link com.softslate.commerce.businessobjects.product.Discount} either is
	 * not a coupon, or is a coupon whose code the current user has already
	 * entered in the session.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @return true, if the <code>Discount</code> either is not a coupon, or is
	 *         a coupon whose code the current user has already entered in the
	 *         session.
	 * @throws Exception
	 * @see com.softslate.commerce.businessobjects.core.User#getEnteredCouponCodes()
	 */
	public boolean isCouponOk(Discount discount) throws Exception {

		// If it's a gift certificate, the coupon code must be populated
		if ((discount.getItemsAppliesType() != null && discount
				.getItemsAppliesType().equals("[giftCertificate]"))
				&& (discount.getCouponCode() == null || discount
						.getCouponCode().length() == 0)) {
			if (log.isDebugEnabled())
				log.debug("Discount not ok - it's a gift cert but has no "
						+ "coupon code.");
			return false;
		}

		// If a coupon code is present the user must have entered in the code
		if (discount.getCouponCode() != null
				&& discount.getCouponCode().length() > 0) {

			// If so, has the user entered in its code?
			if (getUser().getEnteredCouponCodes() == null) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - no coupon codes entered.");
				return false;
			}

			// Find it, ignoring case for case insensitivity
			Iterator i = getUser().getEnteredCouponCodes().iterator();
			while (i.hasNext()) {
				String couponCode = (String) i.next();
				if (couponCode.equalsIgnoreCase(discount.getCouponCode())) {
					if (log.isDebugEnabled())
						log.debug("Discount  ok - coupon code found: "
								+ couponCode);
					return true;
				}
			}

			// Didn't find it
			if (log.isDebugEnabled())
				log.debug("Discount not ok - coupon code not entered.");
			return false;
		}

		// Discount isn't a coupon or a gift certificate, so it's ok.
		return true;
	}

	/**
	 * Checks that the current user is qualified to use a given
	 * {@link com.softslate.commerce.businessobjects.product.Discount} based on
	 * its settings.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @param previousOrderTotal
	 *            The previous order total of the current user, or null. Some
	 *            discounts are only applicable if the user has previously
	 *            placed orders totalling a certain amount.
	 * @return true, if the current user is qualified to use the
	 *         <code>Discount</code>.
	 * @throws Exception
	 */
	public boolean isUserQualified(Discount discount, Double previousOrderTotal)
			throws Exception {
		// Try to disqualify this discount by who the customer is
		if (discount.getUsersAppliesType() != null && discount.getUsersAppliesType().equals("[allUsers]")) {
			if (log.isDebugEnabled())
				log.debug("Discount applies to all users.");
			return true;
		} else if (discount.getUsersAppliesType() != null && discount.getUsersAppliesType().equals("[loggedIn]")) {
			if (getUser().getCustomer() == null) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - user must be logged in.");
				return false;
			}
		} else if (discount.getUsersAppliesType() != null && discount.getUsersAppliesType().equals("[userNames]")) {
			boolean userNameMatches = false;
			if (discount.getUsersAppliesValue() != null
					&& getUser().getCustomer() != null
					&& getUser().getCustomer().getUserName() != null) {
				String[] names = discount.getUsersAppliesValue().split(",");
				for (int j = 0; j < names.length; j++) {
					String name = names[j].trim();
					if (name.toLowerCase().equals(getUser().getCustomer().getUserName().toLowerCase())) {
						userNameMatches = true;
						break;
					}
				}
			}
			if (!userNameMatches) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - user name does not match.");
				return false;
			}
		} else if (discount.getUsersAppliesType() != null
				&& discount.getUsersAppliesType().equals("[customerStatuses]")) {
			boolean statusMatches = false;
			if (discount.getUsersAppliesValue() != null
					&& getUser().getCustomer() != null
					&& getUser().getCustomer().getStatus() != null) {
				String[] statuses = discount.getUsersAppliesValue().split(",");
				for (int j = 0; j < statuses.length; j++) {
					String status = statuses[j].trim();
					if (status.equals(getUser().getCustomer().getStatus())) {
						statusMatches = true;
						break;
					}
				}
			}
			if (!statusMatches) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - cust. status doesn't match.");
				return false;
			}
		} else if (discount.getUsersAppliesType() != null
				&& discount.getUsersAppliesType()
				.equals("[previousOrderTotal]")) {
			if (getUser().getCustomer() == null) {
				return false;
			} else {
				if (previousOrderTotal == null) {
					previousOrderTotal = getPreviousOrderTotal();
				}

				// Subtract the total of the current order if it is completed
				// (ie, this processing was triggered from the admin area).
				if (previousOrderTotal != null
						&& getUser().getOrder().getCompleted() != null
						&& getUser().getOrder().getTotal() != null) {
					previousOrderTotal = new Double(
							previousOrderTotal.doubleValue()
									- getUser().getOrder().getTotal()
											.doubleValue());
				}
				if (previousOrderTotal == null
						|| previousOrderTotal.doubleValue() < Double
								.parseDouble(discount.getUsersAppliesValue())) {
					if (log.isDebugEnabled())
						log.debug("Discount not ok - prev. orders too low.");
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Checks that the current user's cart contains the required items to use a
	 * given {@link com.softslate.commerce.businessobjects.product.Discount}
	 * based on its settings.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @return true, if the current user is qualified to use the
	 *         <code>Discount</code>.
	 * @throws Exception
	 */
	public boolean isCartItemsOk(Discount discount) throws Exception {
		if (discount.getCartItemsType() == null) {
			if (log.isDebugEnabled())
				log.debug("Discount does not require items in the cart.");
			return true;
		} else if (discount.getCartItemsType().equals("[noItemsRequired]")) {
			if (log.isDebugEnabled())
				log.debug("Discount does not require items in the cart.");
			return true;
		} else if (discount.getCartItemsType().equals("[anyOneItem]")) {
			boolean itemsInCart = false;
			if (discount.getCartItemsValue() != null
					&& getUser().getOrder() != null
					&& getUser().getOrder().firstDelivery() != null
					&& getUser().getOrder().firstDelivery()
							.getOrderItemCollection() != null) {
				String[] codes = discount.getCartItemsValue().split(",");
				for (int j = 0; j < codes.length; j++) {
					String code = codes[j].trim();
					Iterator i = getUser().getOrder().firstDelivery()
							.getOrderItemCollection().iterator();
					while (i.hasNext()) {
						OrderItem oi = (OrderItem) i.next();
						if (code.equals(oi.getCode())) {
							itemsInCart = true;
							break;
						}
					}
					if (itemsInCart) {
						break;
					}
				}
			}
			if (!itemsInCart) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - none of the items ("
							+ discount.getCartItemsValue()
							+ ") are in the cart.");
				return false;
			}
		} else if (discount.getCartItemsType().equals("[allItems]")) {
			boolean allItemsInCart = true;
			if (discount.getCartItemsValue() != null
					&& getUser().getOrder() != null
					&& getUser().getOrder().firstDelivery() != null
					&& getUser().getOrder().firstDelivery()
							.getOrderItemCollection() != null) {
				String[] codes = discount.getCartItemsValue().split(",");
				for (int j = 0; j < codes.length; j++) {
					boolean itemInCart = false;
					String code = codes[j].trim();
					Iterator i = getUser().getOrder().firstDelivery()
							.getOrderItemCollection().iterator();
					while (i.hasNext()) {
						OrderItem oi = (OrderItem) i.next();
						if (code.equals(oi.getCode())) {
							itemInCart = true;
							break;
						}
					}
					if (!itemInCart) {
						allItemsInCart = false;
						break;
					}
				}
			} else {
				allItemsInCart = false;
			}
			if (!allItemsInCart) {
				if (log.isDebugEnabled())
					log.debug("Discount not ok - all of the items: ("
							+ discount.getCartItemsValue()
							+ ") are not in the cart.");
				return false;
			}
		}
		return true;
	}

	/**
	 * Checks that the current user's cart contains the required shipping method
	 * to use a given
	 * {@link com.softslate.commerce.businessobjects.product.Discount} based on
	 * its settings.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @return true, if the current user is qualified to use the
	 *         <code>Discount</code>.
	 * @throws Exception
	 */
	public boolean isShippingMethodOk(Discount discount) throws Exception {
		if (getUser().getOrder() == null || discount.getItemsAppliesType() == null
				|| !discount.getItemsAppliesType().equals("[shipping]")) {
			return true;
		}
		String shippingMethod = getUser().getOrder().getShippingMethod();
		if (shippingMethod == null || discount.getExtra1() == null
				|| shippingMethod.length() == 0
				|| discount.getExtra1().length() == 0) {
			if (log.isInfoEnabled())
				log.info("Discount ok - order's or discount's "
						+ "shipping method not defined.");
			return true;
		}
		String[] discountShipMethods = discount.getExtra1().split(",");
		for (int i = 0; i < discountShipMethods.length; i++) {
			String discountShipMethod = discountShipMethods[i];
			if (shippingMethod.trim().equals(discountShipMethod.trim())) {
				if (log.isInfoEnabled())
					log.info("Discount ok - order's shipping method: "
						+ getUser().getOrder().getShippingMethod().trim()
						+ " matches discount's.");
			return true;
		}
		}
		return false;
	}

	/**
	 * Checks whether the given productCode is among the products that are
	 * required to be in the user's cart for the
	 * {@link com.softslate.commerce.businessobjects.product.Discount} to apply,
	 * based on its settings.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @return true, if the current user is qualified to use the
	 *         <code>Discount</code>.
	 * @throws Exception
	 */
	public boolean isItemWouldAllowDiscount(Discount discount,
			String productCode) throws Exception {
		if (discount.getCartItemsType() == null) {
			if (log.isDebugEnabled())
				log.debug("Discount does not require items in the cart.");
			return false;
		} else if (discount.getCartItemsType().equals("[anyOneItem]")) {
			if (discount.getCartItemsValue() != null) {
				String[] codes = discount.getCartItemsValue().split(",");
				for (int j = 0; j < codes.length; j++) {
					String code = codes[j].trim();
					if (code.equals(productCode)) {
						if (log.isDebugEnabled())
							log.debug("Item would allow discount: "
									+ productCode);
						return true;
					}
				}
			}
		}
		if (log.isDebugEnabled())
			log.debug("Item would not allow discount: " + productCode);
		return false;
	}

	/**
	 * Looks up the total of all of the orders the current user has previously
	 * placed. Used to determine if certain discount are applicable.
	 * 
	 * @return The total of all of the orders the current user has previously
	 *         placed.
	 * @throws Exception
	 */
	public Double getPreviousOrderTotal() throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory()
				.createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.getPreviousOrderTotal(getUser().getCustomer()
				.getCustomerID());
	}

	/**
	 * Looks up the number of times the current user has used a given
	 * {@link com.softslate.commerce.businessobjects.product.Discount}.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to check.
	 * @return An <code>Integer</code> representing the number of times the
	 *         current user has used the <code>Discount</code>.
	 * @throws Exception
	 */
	public Integer getTimesUsedByCustomer(Discount discount) throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory()
				.createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.getTimesUsedByCustomer(discount.getDiscountID(),
				getUser().getCustomer().getCustomerID());
	}

	/**
	 * Calculated the value of a discount as applied to the current user's cart.
	 * 
	 * @param discount
	 *            The <code>Discount</code> to be applied.
	 * @param totalToDiscount
	 *            The total that the discount is to be applied against.
	 *            Discounts are not allowed to be more than the amount they are
	 *            being applied against.
	 * @param quantity
	 *            The total number of items in the user's cart, or the quantity
	 *            of an <ocde>OrderItem</code> in the user's cart. Some
	 *            discounts are applied according to
	 *            {@link com.softslate.commerce.businessobjects.product.DiscountRange}
	 *            that corresponds to the total number of items in the user's
	 *            cart (in the case of global discounts), or the quantity of an
	 *            <code>OrderItem</code> correponding to a <code>SKU</code>
	 *            associated with the discount (in the case of <code>SKU</code>
	 *            discounts).
	 * @return A <code>Map</code> with information about the discount, and a key
	 *         named "discountTotal", holding the calculated value of the
	 *         discount.
	 */
	public Map calculateDiscount(Discount discount, Double totalToDiscount,
			Integer quantity) {
		Map results = new HashMap();

		// Subtract the value of gift certificates being purchased
		totalToDiscount = excludeGiftCertificateProducts(discount,
				totalToDiscount, quantity);

		// Skip it if the only items were gift certificate items
		if (!discount.getItemsAppliesType().equals("[shipping]")
				&& totalToDiscount.doubleValue() == 0) {
			return new HashMap();
		}
		Double discountTotal = null;
		if (discount.getRewardType().equals("[percentage]")
				&& discount.getReward() != null) {
			// Percentage discount
			discountTotal = new Double(
					(discount.getReward().doubleValue() * totalToDiscount
							.doubleValue()) / 100);
			if (log.isDebugEnabled())
				log.debug("Percentage discount: " + discount.getReward());
			results.put("reward", discount.getReward());
			results.put("rewardType", discount.getRewardType());
		} else if (discount.getRewardType().equals("[amount]")
				&& discount.getReward() != null) {
			// Flat amount discount - if it's a sku discount, multiple by the
			// quantity
			if (discount.getItemsAppliesType().equals("[skus]")) {
				discountTotal = new Double(discount.getReward().doubleValue()
						* quantity.intValue());
			} else {
				discountTotal = discount.getReward();
			}
			// Now make sure it's not more than the total itself
			if (discountTotal.doubleValue() > totalToDiscount.doubleValue()) {
				discountTotal = totalToDiscount;
			}
			if (log.isDebugEnabled())
				log.debug("Amount discount: " + discount.getReward());
			results.put("reward", discount.getReward());
			results.put("rewardType", discount.getRewardType());
		} else if (discount.getRewardType().equals("[itemsForFree]")
				&& discount.getReward() != null
				&& discount.getItemsAppliesType().equals("[skus]")) {
			// How much is each item?
			double priceEach = totalToDiscount.doubleValue()
					/ quantity.intValue();
			// How many items are free?
			int itemsForFree = Math.round(discount.getReward().floatValue());
			if (quantity.intValue() < itemsForFree) {
				itemsForFree = quantity.intValue();
			}
			// Multiply
			discountTotal = new Double(itemsForFree * priceEach);
			if (discountTotal.doubleValue() > totalToDiscount.doubleValue()) {
				discountTotal = totalToDiscount;
			}
			if (log.isDebugEnabled())
				log.debug("Items for free discount: " + discount.getReward());
			results.put("reward", discount.getReward());
			results.put("rewardType", discount.getRewardType());
		} else {
			// Discount determined by the discount range the total or quantity
			// falls in
			if (discount.getDiscountRanges() != null) {

				// The price ranges for shipping discounts should look at the
				// order's subtotal
				Double rangeCheckTotal = totalToDiscount;
				if (discount.getItemsAppliesType().equals("[shipping]")
						&& getUser().getOrder().getSubtotal() != null) {
					rangeCheckTotal = getUser().getOrder().getSubtotal();
				}
				Iterator i = discount.getDiscountRanges().iterator();
				while (i.hasNext()) {
					DiscountRange dr = (DiscountRange) i.next();
					// If this the right range? (massive conditional)
					if ((discount.getRewardType().equals("[rangeByQuantity]")
							&& quantity.intValue() >= dr.getFloor()
									.doubleValue() && (quantity.intValue() <= dr
							.getCeiling().doubleValue() || dr.getCeiling()
							.doubleValue() == 0))
							|| (discount.getRewardType().equals(
									"[rangeByPrice]")
									&& rangeCheckTotal.doubleValue() >= dr
											.getFloor().doubleValue() && (rangeCheckTotal
									.doubleValue() <= dr.getCeiling()
									.doubleValue() || dr.getCeiling()
									.doubleValue() == 0))) {
						// It's the right range, so calculate the discount
						if (dr.getRewardType().equals("[percentage]")
								&& dr.getReward() != null) {
							discountTotal = new Double(
									(dr.getReward().doubleValue() * totalToDiscount
											.doubleValue()) / 100);
							if (log.isDebugEnabled())
								log.debug("Percentage discount by range: "
										+ dr.getReward());
						} else if (dr.getRewardType().equals("[amount]")
								&& dr.getReward() != null) {
							if (discount.getItemsAppliesType().equals("[skus]")) {
								discountTotal = new Double(dr.getReward()
										.doubleValue() * quantity.intValue());
							} else {
								discountTotal = dr.getReward();
							}
							if (discountTotal.doubleValue() > totalToDiscount
									.doubleValue()) {
								discountTotal = totalToDiscount;
							}
							if (log.isDebugEnabled())
								log.debug("Amount discount by range: "
										+ dr.getReward());
						} else if (dr.getRewardType().equals("[itemsForFree]")
								&& dr.getReward() != null
								&& discount.getItemsAppliesType().equals(
										"[skus]")) {
							// How much is each item?
							double priceEach = totalToDiscount.doubleValue()
									/ quantity.intValue();
							// How many items are free?
							int itemsForFree = Math.round(dr.getReward()
									.floatValue());
							if (quantity.intValue() < itemsForFree) {
								itemsForFree = quantity.intValue();
							}
							// Multiply
							discountTotal = new Double(itemsForFree * priceEach);
							if (discountTotal.doubleValue() > totalToDiscount
									.doubleValue()) {
								discountTotal = totalToDiscount;
							}
							if (log.isDebugEnabled())
								log.debug("Items for free discount by range: "
										+ dr.getReward());
						}
						results.put("floor", dr.getFloor());
						results.put("ceiling", dr.getCeiling());
						results.put("reward", dr.getReward());
						results.put("rewardType", dr.getRewardType());
						break;
					}
				}
			}
		}
		if (discountTotal != null) {
			discountTotal = utils().roundForMoney(discountTotal);
		}
		if (log.isDebugEnabled())
			log.debug("Calculated discount total: " + discountTotal);
		results.put("discountTotal", discountTotal);
		return results;
	}

	/**
	 * When computing a global discount, exclude the value of any gift
	 * certificates that are being purchased if called for by the setting
	 * 'excludeGiftCertificatesFromDiscounts'.
	 * 
	 * @param discount
	 * @param totalToDiscount
	 * @param quantity
	 * @return Updated totalToDiscount that does not include the value of Gift
	 *         Certificates
	 */
	public Double excludeGiftCertificateProducts(Discount discount,
			Double totalToDiscount, Integer quantity) {
		String excludeGiftCertificatesFromDiscounts = (String) getSettings()
				.getValue("excludeGiftCertificatesFromDiscounts");
		String giftCertificateAmountAttributeCode = (String) getSettings()
				.getValue("giftCertificateAmountAttributeCode");
		if (giftCertificateAmountAttributeCode == null || giftCertificateAmountAttributeCode.length() == 0) {
			giftCertificateAmountAttributeCode = "GiftCertificateAmount";
		}
		if (discount.getItemsAppliesType().equals("[global]")
				&& excludeGiftCertificatesFromDiscounts != null
				&& excludeGiftCertificatesFromDiscounts.equals("1")) {
			Iterator i = getUser().getOrder().firstDelivery()
					.getOrderItemCollection().iterator();
			while (i.hasNext()) {
				OrderItem oi = (OrderItem) i.next();
				if (utils().doesOrderItemHaveAttributeCode(oi,
						giftCertificateAmountAttributeCode)) {
					totalToDiscount = new Double(totalToDiscount.doubleValue()
							- oi.getTotal().doubleValue());
					totalToDiscount = new Double(Math.rint(totalToDiscount
							.doubleValue() * 100) / 100);
					if (log.isInfoEnabled())
						log.info("When calculating discount "
								+ discount.getCode()
								+ ", subtracting the value of the gift certificate item "
								+ oi.getCode()
								+ " from the totalToDiscount, which is now: "
								+ totalToDiscount);
				}
			}
		}
		return totalToDiscount;
	}

	/**
	 * Subtracts a discount total from the current user's order.
	 * 
	 * @param discountTotal
	 *            The discount total to subtract.
	 */
	public void updateOrderAndDeliveryTotals(Double discountTotal) {
		// Subtract the discount total from the order's and first
		// delivery's subtotals
		getUser().getOrder().setSubtotal(
				new Double(getUser().getOrder().getSubtotal().doubleValue()
						- discountTotal.doubleValue()));
		getUser()
				.getOrder()
				.firstDelivery()
				.setSubtotal(
						new Double(getUser().getOrder().firstDelivery()
								.getSubtotal().doubleValue()
								- discountTotal.doubleValue()));
		// Redo the formatting
		getUser()
				.getOrder()
				.firstDelivery()
				.addFormatting(getUser().getLocale(),
						(String) getSettings().getValue("currencyCode"));
		getUser().getOrder().addFormatting(getUser().getLocale(),
				(String) getSettings().getValue("currencyCode"));
	}

	/**
	 * Subtracts a gift certificate total from the current user's order's total
	 * (rather than its subtotal as is the case for other types of discounts).
	 * 
	 * @param discountTotal
	 *            The discount total to subtract.
	 */
	public void updateGiftCertOrderAndDeliveryTotals(Double discountTotal) {
		// Gift certificates apply against the total rather than subtotal
		if (getUser().getOrder().getTotal() != null) {
			double newOrderTotal = utils().roundForMoney(
					getUser().getOrder().getTotal().doubleValue()
							- discountTotal.doubleValue()).doubleValue();
			getUser().getOrder().setTotal(new Double(newOrderTotal));
			double newDeliveryTotal = utils().roundForMoney(
					getUser().getOrder().firstDelivery().getTotal()
							.doubleValue()
							- discountTotal.doubleValue()).doubleValue();
			getUser().getOrder().firstDelivery()
					.setTotal(new Double(newDeliveryTotal));
			// Redo the formatting
			getUser()
					.getOrder()
					.firstDelivery()
					.addFormatting(getUser().getLocale(),
							(String) getSettings().getValue("currencyCode"));
			getUser().getOrder().addFormatting(getUser().getLocale(),
					(String) getSettings().getValue("currencyCode"));
		}
	}

	public void addSkuDiscount(Collection skuDiscountsToApply,
			Discount discount, OrderItem oi) {

		// Add the order item to the discount if we are already
		// calculating this order item from a different line item
		Iterator i = skuDiscountsToApply.iterator();
		boolean found = false;
		while (i.hasNext()) {
			Discount ad = (Discount) i.next();
			if (ad.getDiscountID() == discount.getDiscountID()) {
				found = true;
				if (log.isDebugEnabled())
					log.debug("Adding order item to existing discount: "
							+ discount.getCode());
				addOrderItemToSkuDiscount(ad, oi);
				break;
			}
		}

		// Add the discount and the oi if we are not tracking it yet
		if (!found) {
			if (log.isDebugEnabled())
				log.debug("Adding order item to new discount: "
						+ discount.getCode());
			addOrderItemToSkuDiscount(discount, oi);
			skuDiscountsToApply.add(discount);
		}
	}

	public void addOrderItemToSkuDiscount(Discount ad, OrderItem oi) {
		if (ad.getOrderItems() == null) {
			ad.setOrderItems(new ArrayList());
		}
		Iterator i = ad.getOrderItems().iterator();
		boolean found = false;
		// Make sure we dont add an oi that we already have
		if (oi != null) {
			while (i.hasNext()) {
				OrderItem o = (OrderItem) i.next();
				if (o.equals(oi)) {
					found = true;
					break;
				}
			}
			if (!found) {
				ad.getOrderItems().add(oi);
				ad.setTotalQuantity(new Integer(ad.getTotalQuantity()
						.intValue() + oi.getQuantity()));
				ad.setTotalAmount(new Double(ad.getTotalAmount().doubleValue()
						+ oi.getTotal().doubleValue()));
				ad.setTotalWeight(new Double(ad.getTotalWeight().doubleValue()
						+ oi.getWeight().doubleValue()));
			}
		}
	}

	public Map defineDiscountSettings(Product product,
			Collection attributeSKUCollection, Collection storeDiscounts)
			throws Exception {

		// Define displayDiscounts
		boolean displayDiscounts = getSettings().getValue("applyDiscounts")
				.equals("1")
				&& ((product.getSkuMap() != null && product.getSkuMap().size() > 0) || (attributeSKUCollection != null
						&& !attributeSKUCollection.isEmpty() && product.getProductAttributes() != null));

		// Define displayProductDiscount and productDiscounts
		// The productDiscount is the first discount found associated with the
		// product SKU that is of type percentage or amount, such that it can be
		// displayed next to the price. productDiscounts is a Collection of
		// *all* the discounts found for the product SKU.
		boolean displayProductDiscount = false;
		Discount productDiscount = null;
		Collection productDiscounts = new ArrayList();
		Double productDiscountAmount = new Double(0);
		double discAmt = 0;
		Integer productDiscountNumberFree = new Integer(0);
		int discNumFree = 0;
		Map productSkuMap = product.getSkuMap();
		SKU productSku = null;
		String prodSkuKey = product.getProductID() + "|";
		if (productSkuMap != null) {
			productSku = (SKU) product.getSkuMap().get(prodSkuKey);
		}
		if (displayDiscounts && productSku != null
				&& productSku.getSkuDiscounts() != null) {
			Iterator i = productSku.getSkuDiscounts().iterator();
			while (i.hasNext()) {
				SKUDiscount sd = (SKUDiscount) i.next();
				boolean rightRewardType = sd.getDiscount().getRewardType() != null
						&& (sd.getDiscount().getRewardType()
								.equals("[percentage]")
								|| sd.getDiscount().getRewardType()
										.equals("[amount]") || sd.getDiscount()
								.getRewardType().equals("[itemsForFree]"));
				boolean rightItemsType = sd.getDiscount().getItemsAppliesType() != null
						&& sd.getDiscount().getItemsAppliesType()
								.equals("[skus]");
				boolean discountIsOk = rightRewardType && rightItemsType
						&& isDiscountOk(sd.getDiscount(), null);
				if (discountIsOk) {
					displayProductDiscount = true;
					productDiscounts.add(sd.getDiscount());
				}
			}
		}

		// Define productDiscount (legacy) productDiscountExpirationDate and
		// productDiscountAmount.
		// productDiscount is simply the first discount in productDiscounts.
		// productDiscountExpirationDate is the earliest expiration date.
		// Start by sorting the productDiscounts according to their
		// discountOrder field. Then add up the amount of each discount to get
		// the total productDiscountAmount. If any of the discounts here are
		// "itemsForFree" discounts, populate productDiscountNumberFree with the
		// number they get for free.
		Date productDiscountExpiration = null;
		String productDiscountExpirationDate = null;
		if (displayDiscounts && productDiscounts != null
				&& product.getUnitPrice() != null) {
			Collections.sort((List) productDiscounts, new DiscountComparator());
			Iterator i = productDiscounts.iterator();
			while (i.hasNext()) {
				Discount d = (Discount) i.next();
				if (productDiscount == null) {
					productDiscount = d;
				}

				// Set the productDiscountExpirationDate
				if (d.getExpirationDate() != null
						&& d.getExpirationDate().length() > 0) {
					try {
						SimpleDateFormat formatter = new SimpleDateFormat(
								"yyyy-MM-dd HH:mm:ss:SSS", Locale.getDefault());
						Date thisDate = formatter.parse(d.getExpirationDate());
						if (productDiscountExpiration == null
								|| thisDate.before(productDiscountExpiration)) {
							productDiscountExpiration = thisDate;
							productDiscountExpirationDate = d
									.getExpirationDate();
						}
					} catch (Exception e) {
						if (log.isWarnEnabled())
							log.warn("Error parsing discount expiration date "
									+ "for discount: " + d.getCode());
					}
				}

				// Only figure in percentage and amount discounts. Everything
				// else depends on what the customer selects
				if (d.getRewardType() != null
						&& d.getRewardType().equals("[percentage]")) {
					double thisDiscAmt = product.getUnitPrice().doubleValue()
							* (d.getReward().doubleValue() / 100);
					discAmt = discAmt + thisDiscAmt;
				} else if (d.getRewardType() != null
						&& d.getRewardType().equals("[amount]")) {
					discAmt = discAmt + d.getReward().doubleValue();
				} else if (d.getRewardType() != null
						&& d.getRewardType().equals("[itemsForFree]")
						&& discNumFree < d.getReward().intValue()) {
					discNumFree = d.getReward().intValue();
				}
			}
			if (discAmt > product.getUnitPrice().doubleValue()) {
				discAmt = product.getUnitPrice().doubleValue();
			}
			productDiscountAmount = new Double(discAmt);
			productDiscountNumberFree = new Integer(discNumFree);
		}

		// Define displaySkuDiscounts and skuDiscountsToDisplay
		// Create a collection of skus with discounts that should be displayed
		// by discount.jsp
		Collection skuDiscountsToDisplay = new LinkedHashSet();
		Map skuDiscountMap = new HashMap();
		boolean displaySkuDiscounts = false;
		if (displayDiscounts && product.getSkus() != null) {
			Iterator i = product.getSkus().iterator();
			while (i.hasNext()) {
				SKU s = (SKU) i.next();
				if (s.getSkuDiscounts() != null) {
					Iterator j = s.getSkuDiscounts().iterator();
					while (j.hasNext()) {
						SKUDiscount sd = (SKUDiscount) j.next();
						boolean rightItemsType = sd.getDiscount()
								.getItemsAppliesType() != null
								&& sd.getDiscount().getItemsAppliesType()
										.equals("[skus]");
						boolean discountIsOk = sd.getDiscount().getIsActive()
								&& rightItemsType
								&& isDiscountOk(sd.getDiscount(), null);
						if (discountIsOk) {
							Map discountAndSku = new HashMap();
							discountAndSku.put("sku", s);
							discountAndSku.put("discount", sd.getDiscount());
							skuDiscountsToDisplay.add(discountAndSku);
							if (skuDiscountMap.get(new Integer(s.getSkuID())) == null) {
								skuDiscountMap.put(new Integer(s.getSkuID()),
										sd.getDiscount());
							}
							displaySkuDiscounts = true;
						}
					}
				}
			}
		}
		if (displayDiscounts && attributeSKUCollection != null && !attributeSKUCollection.isEmpty()
				&& product.getProductAttributes() != null) {
			Iterator i = product.getProductAttributes().iterator();
			while (i.hasNext()) {
				ProductAttribute pa = (ProductAttribute) i.next();
				Iterator j = attributeSKUCollection.iterator();
				while (j.hasNext()) {
					SKU s = (SKU) j.next();
					if (pa.getAttributeID() == s.getAttributeID1().intValue()) {
						if (s.getSkuDiscounts() != null) {
							Iterator k = s.getSkuDiscounts().iterator();
							while (k.hasNext()) {
								SKUDiscount sd = (SKUDiscount) k.next();
								boolean rightItemsType = sd.getDiscount().getItemsAppliesType() != null
										&& sd.getDiscount().getItemsAppliesType().equals("[skus]");
								boolean discountIsOk = sd.getDiscount().getIsActive() && rightItemsType
										&& isDiscountOk(sd.getDiscount(), null);
								if (discountIsOk) {
									Map discountAndSku = new HashMap();
									discountAndSku.put("sku", s);
									discountAndSku.put("discount", sd.getDiscount());
									skuDiscountsToDisplay.add(discountAndSku);
									if (skuDiscountMap.get(new Integer(s.getSkuID())) == null) {
										skuDiscountMap.put(new Integer(s.getSkuID()),
												sd.getDiscount());
									}
									displaySkuDiscounts = true;
								}
							}
						}
					}
				}
			}
		}

		// Define displayPromotions
		// Discounts that would be applied if this product were ordered
		boolean displayPromotions = false;
		if (storeDiscounts != null) {
			Iterator i = storeDiscounts.iterator();
			while (i.hasNext()) {
				Discount d = (Discount) i.next();
				if (d.getIsActive()
						&& isDiscountOkIfItemInCart(d, null, product.getCode())) {
					displayPromotions = true;
					break;
				}
			}
		}

		Map results = new HashMap();
		results.put("displayDiscounts", new Boolean(displayDiscounts));
		results.put("displayProductDiscount", new Boolean(
				displayProductDiscount));
		results.put("productDiscount", productDiscount);
		results.put("productDiscounts", productDiscounts);
		results.put("productDiscountAmount", productDiscountAmount);
		results.put("productDiscountNumberFree", productDiscountNumberFree);
		results.put("productDiscountExpirationDate",
				productDiscountExpirationDate);
		results.put("displaySkuDiscounts", new Boolean(displaySkuDiscounts));
		results.put("skuDiscountsToDisplay", skuDiscountsToDisplay);
		results.put("skuDiscountMap", skuDiscountMap);
		results.put("displayPromotions", new Boolean(displayPromotions));
		if (log.isDebugEnabled())
			log.debug("Discount settings for " + product.getCode() + ": "
					+ results);
		return results;
	}
}
