package com.softslate.commerce.businessobjects.payment;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;

import net.authorize.Environment;
import net.authorize.api.contract.v1.ArrayOfSetting;
import net.authorize.api.contract.v1.CreateTransactionRequest;
import net.authorize.api.contract.v1.CreateTransactionResponse;
import net.authorize.api.contract.v1.CreditCardType;
import net.authorize.api.contract.v1.CustomerAddressType;
import net.authorize.api.contract.v1.CustomerDataType;
import net.authorize.api.contract.v1.ExtendedAmountType;
import net.authorize.api.contract.v1.MerchantAuthenticationType;
import net.authorize.api.contract.v1.MessageTypeEnum;
import net.authorize.api.contract.v1.NameAndAddressType;
import net.authorize.api.contract.v1.OpaqueDataType;
import net.authorize.api.contract.v1.OrderType;
import net.authorize.api.contract.v1.PaymentType;
import net.authorize.api.contract.v1.SettingType;
import net.authorize.api.contract.v1.TransactionRequestType;
import net.authorize.api.contract.v1.TransactionResponse;
import net.authorize.api.contract.v1.TransactionTypeEnum;
import net.authorize.api.controller.CreateTransactionController;
import net.authorize.api.controller.base.ApiOperationBase;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.daos.payment.PaymentDAO;

/**
 * Processes a payment in the store for a given user's cart. This implementation processes a request through
 * Authorize.net's AIM integration method.
 * <p>
 * If the transaction is declined a result code of 10000 is returned, if it is successful a result code of 0 is
 * returned, and otherwise the result code sent back from Authorize.net is returned.
 * 
 * @author David Tobey
 * @see com.softslate.commerce.daos.payment.PaymentDAO
 */
public class AuthorizeNetAIMProcessor extends BasePaymentProcessor implements PaymentProcessor {

	static Log log = LogFactory.getLog(AuthorizeNetAIMProcessor.class);

	public Map process(Map parameters) throws Exception {
		log.info("Processing payment.");

		// Transaction type from settings
		String authorizeNetTxType = (String) getSettings().getValue("authorizeNetTxType");
		if (authorizeNetTxType.equals("AUTH_CAPTURE")) {
			parameters.put("TransactionType", TransactionTypeEnum.AUTH_CAPTURE_TRANSACTION.value());
		} else {
			parameters.put("TransactionType", TransactionTypeEnum.AUTH_ONLY_TRANSACTION.value());
		}
		return processTransaction(parameters);
	}

	public Map processAuthOnly(Map parameters) throws Exception {
		log.info("Processing auth capture.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.AUTH_ONLY_TRANSACTION.value());
		return processTransaction(parameters);
	}

	public Map processAuthCapture(Map parameters) throws Exception {
		log.info("Processing auth capture.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.AUTH_CAPTURE_TRANSACTION.value());
		return processTransaction(parameters);
	}

	public Map processPriorAuthCapture(Map parameters) throws Exception {
		log.info("Processing prior auth capture.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.PRIOR_AUTH_CAPTURE_TRANSACTION.value());
		return processTransaction(parameters);
	}

	public Map processVoid(Map parameters) throws Exception {
		log.info("Processing void.");

		// transaction fields
		parameters.put("TransactionType", TransactionTypeEnum.VOID_TRANSACTION.value());
		return processTransaction(parameters);
	}

	protected Map processTransaction(Map parameters) throws Exception {
		log.info("Processing transaction.");

		// Test mode
		ApiOperationBase.setEnvironment(configureEnvironment(parameters));

		// Authentication
		ApiOperationBase.setMerchantAuthentication(configureMerchantAuthentication(parameters));

		// Instantiate the transaction request
		TransactionRequestType txnRequest = initTransactionRequest(parameters);

		// Populate the payment data
		txnRequest.setPayment(configurePayment(parameters));

		// Total amount, tax and shipping
		configureTotals(txnRequest, parameters);

		// Invoice and customer reference fields
		configureReferenceFields(txnRequest, parameters);

		// Transaction settings such as email control fields and the duplicate transaction window
		configureTransactionSettings(txnRequest, parameters);

		// customer name and billing address fields
		if (getUser() != null && getUser().getOrder() != null) {
			txnRequest.setBillTo(configureBillingAddress(parameters));
		}

		// customer shipping fields
		if (getUser() != null && getUser().getOrder() != null && getUser().getOrder().firstDelivery() != null) {
			txnRequest.setShipTo(configureShippingAddress(parameters));
		}

		// Make the API Request
		CreateTransactionRequest apiRequest = new CreateTransactionRequest();
		apiRequest.setTransactionRequest(txnRequest);
		CreateTransactionController controller = new CreateTransactionController(apiRequest);
		controller.execute();
		CreateTransactionResponse response = controller.getApiResponse();

		// process the response
		return processResponse(response, parameters);
	}

	protected Environment configureEnvironment(Map parameters) {
		String testMode = getSettings().getValue("authorizeNetTestRequest");
		if (testMode != null && testMode.equals("1")) {
			return Environment.SANDBOX;
		} else {
			return Environment.PRODUCTION;
		}
	}

	protected MerchantAuthenticationType configureMerchantAuthentication(Map parameters) {
		MerchantAuthenticationType merchantAuthenticationType = new MerchantAuthenticationType();
		merchantAuthenticationType.setName(getSettings().getValue("authorizeNetLoginID"));
		merchantAuthenticationType.setTransactionKey(getSettings().getValue("authorizeNetTxKey"));
		return merchantAuthenticationType;
	}

	protected TransactionRequestType initTransactionRequest(Map parameters) {
		TransactionRequestType txnRequest = new TransactionRequestType();
		txnRequest.setCurrencyCode(getSettings().getValue("currencyCode"));
		txnRequest.setTransactionType((String) parameters.get("TransactionType"));

		// For voids or other transactions requiring a reference to a prior transaction
		txnRequest.setRefTransId((String) parameters.get("x_trans_id"));
		return txnRequest;
	}

	protected PaymentType configurePayment(Map parameters) {
		PaymentType paymentType = new PaymentType();
		if ("1".equals(getSettings().getValue("authorizeNetEnableAcceptJS"))) {

			// Use opaque data from Accept.js
			OpaqueDataType opaqueData = new OpaqueDataType();
			opaqueData.setDataDescriptor((String) parameters.get("acceptJsDataDesc"));
			opaqueData.setDataValue((String) parameters.get("acceptJsDataValue"));
			paymentType.setOpaqueData(opaqueData);
		} else {

			// Use credit card data if Accept.js is not enabled
			CreditCardType creditCard = new CreditCardType();
			creditCard.setCardNumber((String) parameters.get("decryptedNumber"));
			if (parameters.get("creditCardExpirationMonth") != null
					&& parameters.get("creditCardExpirationYear") != null) {
				creditCard.setExpirationDate(parameters.get("creditCardExpirationMonth").toString()
						+ parameters.get("creditCardExpirationYear").toString());
			}
			if (parameters.get("creditCardSecurityCode") != null) {
				creditCard.setCardCode((String) parameters.get("creditCardSecurityCode"));
			}
			paymentType.setCreditCard(creditCard);
		}
		return paymentType;
	}

	protected void configureTotals(TransactionRequestType txnRequest, Map parameters) {
		if (getUser() != null && getUser().getOrder() != null && parameters.get("x_amount") == null
				&& getUser().getOrder().getTotal() != null) {
			Double total = getUser().getOrder().getTotal();
			total = utils().roundForMoney(total);
			parameters.put("amount", new BigDecimal(total).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setAmount(new BigDecimal(total).setScale(2, RoundingMode.HALF_UP));
		} else if (parameters.get("x_amount") != null) {
			parameters.put("amount",
					new BigDecimal((String) parameters.get("x_amount")).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setAmount(new BigDecimal((String) parameters.get("x_amount")).setScale(2, RoundingMode.HALF_UP));
		}
		ExtendedAmountType freight = new ExtendedAmountType();
		freight.setName("Freight");
		if (getUser() != null && getUser().getOrder() != null && parameters.get("x_freight") == null
				&& getUser().getOrder().getShipping() != null) {
			Double shipping = getUser().getOrder().getShipping();
			shipping = utils().roundForMoney(shipping);
			freight.setAmount(new BigDecimal(shipping).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setShipping(freight);
		} else if (parameters.get("x_freight") != null) {
			freight.setAmount(new BigDecimal((String) parameters.get("x_freight")).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setShipping(freight);
		}
		ExtendedAmountType tax = new ExtendedAmountType();
		tax.setName("tax");
		if (getUser() != null && getUser().getOrder() != null && parameters.get("x_tax") == null
				&& getUser().getOrder().getTax() != null) {
			Double taxAmount = getUser().getOrder().getTax();
			taxAmount = utils().roundForMoney(taxAmount);
			tax.setAmount(new BigDecimal(taxAmount).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setTax(tax);
		} else if (parameters.get("x_tax") != null) {
			tax.setAmount(new BigDecimal((String) parameters.get("x_tax")).setScale(2, RoundingMode.HALF_UP));
			txnRequest.setTax(tax);
		}
	}

	protected void configureReferenceFields(TransactionRequestType txnRequest, Map parameters) {
		txnRequest.setPoNumber((String) parameters.get("poNumber"));
		OrderType order = new OrderType();
		order.setInvoiceNumber((String) parameters.get("invoiceNumber"));
		order.setDescription((String) parameters.get("description"));
		txnRequest.setOrder(order);
		CustomerDataType customer = new CustomerDataType();
		customer.setId((String) parameters.get("customerID"));
		txnRequest.setCustomer(customer);
		txnRequest.setCustomerIP((String) parameters.get("ipAddress"));
	}

	protected void configureTransactionSettings(TransactionRequestType txnRequest, Map parameters) {
		ArrayOfSetting settings = new ArrayOfSetting();
		SettingType emailCustomer = new SettingType();
		emailCustomer.setSettingName("emailCustomer");
		String emailCust = (String) getSettings().getValue("authorizeNetEmailCustomer");
		if (emailCust != null && emailCust.equals("1")) {
			emailCustomer.setSettingValue("TRUE");
		} else {
			emailCustomer.setSettingValue("FALSE");
		}
		settings.getSetting().add(emailCustomer);
		txnRequest.setTransactionSettings(settings);

		// duplicateWindow
		SettingType duplicateWindow = new SettingType();
		duplicateWindow.setSettingName("duplicateWindow");
		duplicateWindow.setSettingValue(getSettings().getValue("authorizeNetDupeTxTimeout"));
		settings.getSetting().add(duplicateWindow);
	}

	protected CustomerAddressType configureBillingAddress(Map parameters) {
		CustomerAddressType billTo = new CustomerAddressType();
		if (getUser().getOrder().getFirstName() != null) {
			billTo.setFirstName(getUser().getOrder().getFirstName());
		}
		if (getUser().getOrder().getLastName() != null) {
			billTo.setLastName(getUser().getOrder().getLastName());
		}
		if (getUser().getOrder().getOrganization() != null) {
			billTo.setCompany(getUser().getOrder().getOrganization());
		}
		if (getUser().getOrder().getAddress1() != null) {
			billTo.setAddress(getUser().getOrder().getAddress1());
		}
		if (getUser().getOrder().getAddress2() != null) {
			billTo.setAddress(getUser().getOrder().getAddress1() + " " + getUser().getOrder().getAddress2());
		}
		if (getUser().getOrder().getCity() != null) {
			billTo.setCity(getUser().getOrder().getCity());
		}
		if (getUser().getOrder().getState() != null) {
			billTo.setState(getUser().getOrder().getState());
		}
		if (getUser().getOrder().getPostalCode() != null) {
			billTo.setZip(getUser().getOrder().getPostalCode());
		}
		if (getUser().getOrder().getCountry() != null) {
			billTo.setCountry(getUser().getOrder().getCountry());
		}
		if (getUser().getOrder().getPhone1() != null) {
			billTo.setPhoneNumber(getUser().getOrder().getPhone1());
		}
		if (getUser().getOrder().getPhone2() != null) {
			billTo.setFaxNumber(getUser().getOrder().getPhone2());
		}
		if (getUser().getOrder().getEmail1() != null) {
			billTo.setEmail(getUser().getOrder().getEmail1());
		}
		return billTo;
	}

	protected NameAndAddressType configureShippingAddress(Map parameters) {
		CustomerAddressType shipTo = new CustomerAddressType();
		if (getUser().getOrder().firstDelivery().getFirstName() != null) {
			shipTo.setFirstName(getUser().getOrder().firstDelivery().getFirstName());
		}
		if (getUser().getOrder().firstDelivery().getLastName() != null) {
			shipTo.setLastName(getUser().getOrder().firstDelivery().getLastName());
		}
		if (getUser().getOrder().firstDelivery().getOrganization() != null) {
			shipTo.setCompany(getUser().getOrder().firstDelivery().getOrganization());
		}
		if (getUser().getOrder().firstDelivery().getAddress1() != null) {
			shipTo.setAddress(getUser().getOrder().firstDelivery().getAddress1());
		}
		if (getUser().getOrder().firstDelivery().getAddress2() != null) {
			shipTo.setAddress(getUser().getOrder().firstDelivery().getAddress1() + " "
					+ getUser().getOrder().firstDelivery().getAddress2());
		}
		if (getUser().getOrder().firstDelivery().getCity() != null) {
			shipTo.setCity(getUser().getOrder().firstDelivery().getCity());
		}
		if (getUser().getOrder().firstDelivery().getState() != null) {
			shipTo.setState(getUser().getOrder().firstDelivery().getState());
		}
		if (getUser().getOrder().firstDelivery().getPostalCode() != null) {
			shipTo.setZip(getUser().getOrder().firstDelivery().getPostalCode());
		}
		if (getUser().getOrder().firstDelivery().getCountry() != null) {
			shipTo.setCountry(getUser().getOrder().firstDelivery().getCountry());
		}
		if (getUser().getOrder().firstDelivery().getPhone1() != null) {
			shipTo.setPhoneNumber(getUser().getOrder().firstDelivery().getPhone1());
		}
		if (getUser().getOrder().firstDelivery().getPhone2() != null) {
			shipTo.setFaxNumber(getUser().getOrder().firstDelivery().getPhone2());
		}
		if (getUser().getOrder().firstDelivery().getEmail1() != null) {
			shipTo.setEmail(getUser().getOrder().firstDelivery().getEmail1());
		}
		return shipTo;
	}

	protected Map processResponse(CreateTransactionResponse response, Map parameters) throws Exception {
		String resultCode = null;
		String resultMessage = null;
		if (response != null) {

			// Log the overall response and check it for an error (such as an authentication issue)
			if (response.getMessages() != null) {

				// Log it
				MessageTypeEnum responseCode = response.getMessages().getResultCode();
				log.info("Result Code: " + responseCode);
				if (response.getMessages() != null && response.getMessages().getMessage() != null
						&& response.getMessages().getMessage().get(0) != null) {
					resultCode = response.getMessages().getMessage().get(0).getCode();
					resultMessage = response.getMessages().getMessage().get(0).getText();
					log.info("Message Code: " + resultCode);
					log.info("Message Text: " + resultMessage);
				}

				// Check it
				if (responseCode != null && responseCode == MessageTypeEnum.ERROR) {
					return processErrorResponse(parameters, resultCode, resultMessage);
				}
			}

			// Log the transaction response and check it for an error (such as payment declined)
			TransactionResponse transResponse = response.getTransactionResponse();
			if (transResponse != null) {

				// Log it
				log.info("Transaction ID: " + transResponse.getTransId());
				log.info("Response Code: " + transResponse.getResponseCode());
				if (transResponse.getMessages() != null && transResponse.getMessages().getMessage() != null
						&& transResponse.getMessages().getMessage().get(0) != null) {
					resultCode = transResponse.getMessages().getMessage().get(0).getCode();
					resultMessage = transResponse.getMessages().getMessage().get(0).getDescription();
					log.info("Transaction Response Code: " + resultCode);
					log.info("Transaction Response Description: " + resultMessage);
				}
				if (transResponse.getErrors() != null && transResponse.getErrors().getError() != null
						&& transResponse.getErrors().getError().get(0) != null) {
					resultCode = transResponse.getErrors().getError().get(0).getErrorCode();
					resultMessage = transResponse.getErrors().getError().get(0).getErrorText();
					log.info("Transaction Error Code: " + resultCode);
					log.info("Transaction Error Description: " + resultMessage);
				}

				// Check it
				if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("1")) {
					log.info("Payment Approved");
					return processSuccessResponse(parameters, response);
				} else if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("2")) {
					log.info("Payment Declined");

					// The Transaction Error Description is very misleading in this case - tells the user to call
					// Merchant Services. Just tell them the card was declined.
					return processErrorResponse(parameters, resultCode, "Credit Card Declined");
				} else if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("3")) {
					log.info("Payment Error");
					return processErrorResponse(parameters, resultCode, resultMessage);
				} else if (transResponse.getResponseCode() != null && transResponse.getResponseCode().equals("4")) {
					log.info("Held for Review - treating as successful");
					return processSuccessResponse(parameters, response);
				} else {
					log.info("Not a recognized transaction response code - treating as a failure: " + transResponse.getResponseCode());
					return processErrorResponse(parameters, resultCode, resultMessage);
				}
			} else {
				log.info("Null Transaction Response - treating as a failure");
				return processErrorResponse(parameters, resultCode, resultMessage);
			}
		} else {
			log.info("Null Response - treating as a failure");
			return processErrorResponse(parameters, resultCode, resultMessage);
		}
	}

	protected Map processSuccessResponse(Map parameters, CreateTransactionResponse response) throws Exception {
		Map result = new HashMap();
		result.put("resultCode", "0");
		result.put("resultMessage", response.getTransactionResponse().getMessages().getMessage().get(0)
				.getDescription());
		Payment payment = (Payment) getBusinessObjectFactory().createObject("paymentImplementer");
		BeanUtils.copyProperties(payment, parameters);
		payment.setStatus("Received");
		payment.setCreated(formatDateTime(new java.util.Date()));
		payment.setProcessorClassName(this.getClass().getName());
		String creditCardNumberDisplay = null;
		String decryptedNumber = payment.getDecryptedNumber();
		if (decryptedNumber != null) {
			creditCardNumberDisplay = "XXXXXXXXXXXX"
					+ decryptedNumber.substring(decryptedNumber.length() - 4, decryptedNumber.length());
			payment.setCreditCardNumberDisplay(creditCardNumberDisplay);
		}
		payment.setValueFieldLabels("TRANSACTION_TYPE,AMOUNT,CARD_CODE,RESPONSE_CODE,AVS_CODE,AUTHCODE,TRANSACTION_ID,NONCE");
		payment.setValue1((String) parameters.get("TransactionType"));
		if (parameters.get("amount") != null) {
			payment.setValue2(((BigDecimal) parameters.get("amount")).toString());
		}
		payment.setValue3(response.getTransactionResponse().getCavvResultCode());
		payment.setValue4(response.getTransactionResponse().getResponseCode());
		payment.setValue5(response.getTransactionResponse().getAvsResultCode());
		payment.setValue6(response.getTransactionResponse().getAuthCode());
		payment.setValue7(response.getTransactionResponse().getTransId());
		payment.setValue8((String) parameters.get("acceptJsDataValue"));
		payment.setOrder(getUser().getOrder());
		if (getUser().getOrder().getPayments() == null)
			getUser().getOrder().setPayments(new LinkedHashSet());
		getUser().getOrder().getPayments().add(payment);
		PaymentDAO paymentDAO = (PaymentDAO) getDaoFactory().createDAO("paymentDAOImplementer");
		paymentDAO.setPayment(payment);
		paymentDAO.insertPayment(false);
		result.put("newPayment", payment);
		return result;
	}

	protected Map processErrorResponse(Map parameters, String code, String message) throws Exception {
		Map result = new HashMap();
		if (code == null) {
			code = "10000";
		}
		result.put("resultCode", code);
		result.put("resultMessage", message);
		return result;
	}
}
