package com.softslate.commerce.businessobjects.payment;

import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.ApplicationException;
import com.softslate.commerce.daos.payment.PaymentDAO;

/**
 * Processes a payment in the store for a given user's cart. This implementation
 * processes a request through PayPal Payflow Link. Phone numbers and emails
 * are taken out of the request as those are validated by PayPal and may cause
 * inconvenient errors. Use comment1 and comment2 to send emails and phone
 * numbers instead. If the transaction is declined a result code of 10000 is
 * returned, otherwise the result code sent back from PayPal is returned.
 * 
 * @author David Tobey
 * @see com.softslate.commerce.daos.payment.PaymentDAO
 */
public class PayflowLinkProcessor extends BasePaymentProcessor implements
        PaymentProcessor {

    static Log log = LogFactory.getLog(PayflowLinkProcessor.class);

    public Map process(Map parameters) throws Exception {
        if (log.isInfoEnabled())
            log.info("Processing payment.");

        Map result = new HashMap();

        // Prepare the request to PayPal Payflow Link
        String address = "";
        String city = "";
        String state = "";
        String zip = "";
        String country = "";
        String email = "";
        String nameToShip = "";
        String addressToShip = "";
        String cityToShip = "";
        String stateToShip = "";
        String zipToShip = "";
        String countryToShip = "";
        if (getUser().getOrder().getAddress1() != null)
            address = getUser().getOrder().getAddress1();
        if (getUser().getOrder().getAddress2() != null)
            address = address + " " + getUser().getOrder().getAddress2();
        if (getUser().getOrder().getCity() != null)
            city = getUser().getOrder().getCity();
        if (getUser().getOrder().getState() != null)
            state = getUser().getOrder().getState();
        if (getUser().getOrder().getPostalCode() != null)
            zip = getUser().getOrder().getPostalCode();
        if (getUser().getOrder().getCountry() != null)
            country = getUser().getOrder().getCountry();
        if (getUser().getOrder().getEmail1() != null)
            email = getUser().getOrder().getEmail1();
        if (getUser().getOrder().firstDelivery().getFirstName() != null)
            nameToShip = getUser().getOrder().firstDelivery().getFirstName();
        if (getUser().getOrder().firstDelivery().getLastName() != null)
            nameToShip = nameToShip + " "
                    + getUser().getOrder().firstDelivery().getLastName();
        if (getUser().getOrder().firstDelivery().getAddress1() != null)
            addressToShip = getUser().getOrder().firstDelivery().getAddress1();
        if (getUser().getOrder().firstDelivery().getAddress2() != null)
            addressToShip = addressToShip + " "
                    + getUser().getOrder().firstDelivery().getAddress2();
        if (getUser().getOrder().firstDelivery().getCity() != null)
            cityToShip = getUser().getOrder().firstDelivery().getCity();
        if (getUser().getOrder().firstDelivery().getState() != null)
            stateToShip = getUser().getOrder().firstDelivery().getState();
        if (getUser().getOrder().firstDelivery().getPostalCode() != null)
            zipToShip = getUser().getOrder().firstDelivery().getPostalCode();
        if (getUser().getOrder().firstDelivery().getCountry() != null)
            countryToShip = getUser().getOrder().firstDelivery().getCountry();

        Map params = new HashMap();
        params.put("address", address);
        params.put("city", city);
        params.put("state", state);
        params.put("zip", zip);
        params.put("country", country);
        params.put("email", email);
        params.put("nameToShip", nameToShip);
        params.put("addressToShip", addressToShip);
        params.put("cityToShip", cityToShip);
        params.put("stateToShip", stateToShip);
        params.put("zipToShip", zipToShip);
        params.put("countryToShip", countryToShip);
        params.put("amount", getUser().getOrder().getTotal().toString());
        params.put("tax", getUser().getOrder().getTax().toString());
        params.put("shipamount", getUser().getOrder().getShipping().toString());
        String creditCardExpirationYear = (String) parameters
                .get("creditCardExpirationYear");
        params.put("expdate", (String) parameters
                .get("creditCardExpirationMonth")
                + creditCardExpirationYear.substring(2, 4));
        params.put("name", (String) parameters.get("creditCardName"));
        params.put("cardnum", (String) parameters.get("decryptedNumber"));
        params.put("cvv2", (String) parameters.get("creditCardSecurityCode"));
        params.put("user1", (String) parameters.get("creditCardType"));
        params.put("login", (String) getSettings().getValue(
                "payflowLinkUserName"));
        params.put("type", (String) getSettings().getValue("payflowLinkType"));
        params.put("partner", (String) getSettings().getValue(
                "payflowLinkPartner"));
        params.put("method", "CC");
        params.put("orderform", "false");
        params.put("showconfirm", "false");
        params.put("disablereceipt", "false");
        params.put("echodata", "true");
        params.put("description", (String) parameters.get("description"));
        params.put("comment1", (String) parameters.get("comment1"));
        params.put("comment2", (String) parameters.get("comment2"));
        params.put("invoice", (String) parameters.get("invoiceNumber"));
        params.put("ponum", (String) parameters.get("poNumber"));

        // Send the request to Payflow Link
        String postData = prepareRequestData(params, true);
        String urlString = (String) getSettings().getValue("payflowLinkURL");
        String response = sendHTTPPost(urlString, postData, false);

        // Parse the response, looking for PayPal's response variables.
        Map responseMap = new HashMap();
        String regEx = "value ?= ?\"([^\"]*)\"[^<>]*?\"(RESULT|RESPMSG|AVSDATA|AUTHCODE|PNREF)\"";
        Pattern pattern = Pattern.compile(regEx, Pattern.CASE_INSENSITIVE);
        Matcher matcher = pattern.matcher(response);
        while (matcher.find()) {
            String value = matcher.group(1);
            String key = matcher.group(2);
            if (log.isDebugEnabled())
                log.debug("Key: " + key);
            if (log.isDebugEnabled())
                log.debug("Value: " + value);
            responseMap.put(key, value);
        }

        // Handle the response - errors first
        String resultCode = (String) responseMap.get("RESULT");
        if (resultCode == null) {
            if (response.indexOf("Transaction declined") != -1) {
                result.put("resultCode", "10000");
                return result;
            } else if (response.indexOf("Error Page") != -1) {
                String resultMessage = response.replaceAll("<head>.*?</head>",
                        "");
                resultMessage = resultMessage.replaceAll("<.*?>", " ");
                result.put("resultMessage", resultMessage);
                return result;
            } else if (response.indexOf("Entry Errors") != -1) {
                String resultMessage = response.replaceAll("<head>.*?</head>",
                        "");
                resultMessage = resultMessage.replaceAll("<.*?>", " ");
                result.put("resultMessage", "Data entry errors: "
                        + resultMessage);
                return result;
            } else {
                response = response.replaceAll("<head>.*?</head>", "");
                response = response.replaceAll("<.*?>", " ");
                throw new ApplicationException(response);
            }
        } else if (!resultCode.equals("0")) {
            result.put("resultCode", resultCode);
            result.put("resultMessage", (String) responseMap.get("RESPMSG"));
            return result;
        }

        // Successful response
        Payment payment = (Payment) getBusinessObjectFactory().createObject(
                "paymentImplementer");
        BeanUtils.copyProperties(payment, parameters);
        payment.setStatus("Received");
		payment.setAmount(getUser().getOrder().getTotal());
		payment.setCreated(formatDateTime(new java.util.Date()));
        payment.setProcessorClassName(this.getClass().getName());
        String creditCardNumberDisplay = null;
        String decryptedNumber = payment.getDecryptedNumber();
        if (decryptedNumber != null) {
            creditCardNumberDisplay = "XXXXXXXXXXXX"
                    + decryptedNumber.substring(decryptedNumber.length() - 4,
                            decryptedNumber.length());
            payment.setCreditCardNumberDisplay(creditCardNumberDisplay);
        }
        payment
                .setValueFieldLabels(",,,RESULT,AVSDATA,AUTHCODE,PNREF,RESPMSG,");
        payment.setValue4((String) responseMap.get("RESULT"));
        payment.setValue5((String) responseMap.get("AVSDATA"));
        payment.setValue6((String) responseMap.get("AUTHCODE"));
        payment.setValue7((String) responseMap.get("PNREF"));
        payment.setValue8((String) responseMap.get("RESPMSG"));
        payment.setOrder(getUser().getOrder());
        if (getUser().getOrder().getPayments() == null) {
            getUser().getOrder().setPayments(new LinkedHashSet());
        }
        getUser().getOrder().getPayments().add(payment);
        PaymentDAO paymentDAO = (PaymentDAO) getDaoFactory().createDAO(
                "paymentDAOImplementer");
        paymentDAO.setPayment(payment);
        paymentDAO.insertPayment(false);

        result.put("resultCode", "0");
        return result;
    }

}
