package com.softslate.commerce.customer.order;

import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.customer.core.BaseForm;

/**
 * Struts form class for submissions to add items to the current user's cart.
 * Used for requests to "/CartAdd.do" and "/CartAddMultiple.do".
 * <p>
 * The <code>validate</code> method of this class checks the
 * <code>orderProductQuantities</code> property to make sure the user entered
 * a quantity greater than zero for at least one product. It checks the
 * <code>orderProductCodes</code> property to make sure that at least one
 * product code was submitted.
 * <p>
 * It also populates the <code>orderAttributes</code> property with submitted
 * product attributes and their values, and checks the
 * <code>requiredAttributeCodes</code> property to make sure values were
 * submitted for required product attributes.
 * <p>
 * Here is a summary of the request parameters involved in a request to add
 * items to the cart:
 * <ul>
 * <li><code>orderProductQuantities</code>: An array of integers
 * corresponding to the quantities of each item being added to the cart.</li>
 * <li><code>orderProductCodes</code>: An array of product codes
 * corresponding to each item being added to the cart. The order of the product
 * codes must correspond to the order of the quantities in the
 * <code>orderProductQuantities</code> parameter.</li>
 * <li><code>requiredAttributeCodes</code>: An array of required attributes.
 * Each element of the array is a pipe-separated concatenation of the attribute
 * code and the product code, in the form of productCode|attributeCode. Note
 * that the check for required attributes is done in this class based on
 * elements included in the request submission. It is also repeated within
 * {@link com.softslate.commerce.businessobjects.order.CartProcessor#processAddItems(Map)},
 * against the database itself.</li>
 * <li><code>requiredAttributeNames</code>: An array of the names of the
 * required attributes. Each element of the array is the name of the
 * corresponding required attribute in the <code>requiredAttributeCodes</code>
 * property. They are included in the request so that an error message can be
 * displayed to the customer without the need to hit the database.</li>
 * <li>Attributes and their values: Any number of additional request parameters
 * may exist that contain the attribute values selected or entered by the
 * customer. The <em>names</em> of these parameters are a pipe-separated
 * concatenation of the product code and attribute code, in the form of
 * productCode|attributeCode. The values of each parameter contain what the
 * customer entered, in the case of the expandedInput and compactInput attribute
 * types. In the case of the other attribute types, the value consists of the
 * code of the option that the customer selected.</li>
 * </ul>
 * <p>
 * 
 * @author David Tobey
 * @see com.softslate.commerce.customer.order.CartAddAction
 */
public class CartAddForm extends BaseForm {

	private static final long serialVersionUID = 3257848800826700600L;

	static Log log = LogFactory.getLog(CartAddForm.class);

	private String[] orderProductCodes = null;

	public String[] getOrderProductCodes() {
		return (this.orderProductCodes);
	}

	public void setOrderProductCodes(String[] orderProductCodes) {
		this.orderProductCodes = orderProductCodes;
	}

	private String[] requiredAttributeCodes = null;

	public String[] getRequiredAttributeCodes() {
		return (this.requiredAttributeCodes);
	}

	public void setRequiredAttributeCodes(String[] requiredAttributeCodes) {
		this.requiredAttributeCodes = requiredAttributeCodes;
	}

	private String[] requiredAttributeNames = null;

	public String[] getRequiredAttributeNames() {
		return (this.requiredAttributeNames);
	}

	public void setRequiredAttributeNames(String[] requiredAttributeNames) {
		this.requiredAttributeNames = requiredAttributeNames;
	}

	private String[] productNames = null;

	public String[] getProductNames() {
		return productNames;
	}

	public void setProductNames(String[] productNames) {
		this.productNames = productNames;
	}

	private String[] orderProductQuantities = null;

	public String[] getOrderProductQuantities() {
		return (this.orderProductQuantities);
	}

	public void setOrderProductQuantities(String[] orderProductQuantities) {
		this.orderProductQuantities = orderProductQuantities;
	}

	private HashMap orderAttributes = new HashMap();

	public HashMap getOrderAttributes() {
		return (this.orderAttributes);
	}

	public void setOrderAttributes(HashMap orderAttributes) {
		this.orderAttributes = orderAttributes;
	}

	public ActionErrors validate(ActionMapping mapping,
			HttpServletRequest request) {
		if (log.isInfoEnabled())
			log.info("Starting validation.");
		initializeProperties(mapping, request);

		// Check for catalog mode
		try {
			String catalogMode = (String) getSettings().getValue("catalogMode");
			if (catalogMode.equals("1")) {
				getErrors().add(ActionMessages.GLOBAL_MESSAGE,
						new ActionMessage("errors.catalogMode"));
				return getErrors();
			}
		} catch (Exception e) {
		}

		// Check that the user entered a quantity greater than 0 for at least
		// one product.
		validateQuantity();

		// Check that at least one product code was submitted.
		validateProductCode();

		// Validate attributes
		validateAttributes();

		// Check that the cart will not exceed the max items allowed per
		// order
		validateMaxItemsPerOrder();

		if (log.isDebugEnabled())
			log.debug("Finished validation.");
		if (log.isInfoEnabled() && !getErrors().isEmpty())
			log.info("Validation error(s): " + getErrors().toString());
		return getErrors();
	}

	/**
	 * Checks that the user entered a quantity greater than 0 for at least one
	 * product.
	 */
	public void validateQuantity() {

		// Check that the user entered a quantity greater than 0 for at least
		// one product.
		if (!formUtils().stringArrayValuesGreaterThan(
				getOrderProductQuantities(), 0)) {
			String message = getMessage("order",
					"orderForm.productQuantityRequired", null);
			getErrors().add(ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("errors.required", message));
			getRequest().setAttribute("errorProductCode", "");
			if (getOrderProductCodes() != null
					&& getOrderProductCodes().length > 0) {
				getRequest().setAttribute("errorProductCode",
						getOrderProductCodes()[0]);
			}
		}
	}

	/**
	 * Checks that at least one product code was submitted.
	 */
	public void validateProductCode() {

		// Check that at least one product code was submitted.
		if (!formUtils().stringArrayHasAnElement(getOrderProductCodes())) {
			String message = getMessage("order", "orderForm.productRequired",
					null);
			getErrors().add(ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("errors.required", message));
			getRequest().setAttribute("errorProductCode", "");
			if (getOrderProductCodes() != null
					&& getOrderProductCodes().length > 0) {
				getRequest().setAttribute("errorProductCode",
						getOrderProductCodes()[0]);
			}
		}
	}

	/**
	 * Checks that required attributes were submitted, and constructs this
	 * form's <code>orderAttributes</code> property, which holds a Map of the
	 * submitted attributes.
	 */
	public void validateAttributes() {

		// Check that required attributes were submitted
		// This check also performed later against the database
		// Skip if no attributes are required.
		if (getRequiredAttributeCodes() != null
				&& getOrderProductCodes() != null) {

			// Loop through the products with quantities > 0
			for (int i = 0; i < getOrderProductCodes().length; i++) {
				String[] quantity = { getOrderProductQuantities()[i] };
				if (formUtils().stringArrayValuesGreaterThan(quantity, 0)) {

					// Loop through the requiredAttributeCodes - for this
					// product
					for (int j = 0; j < getRequiredAttributeCodes().length; j++) {
						String parameterName = getRequiredAttributeCodes()[j];
						if (parameterName.startsWith(getOrderProductCodes()[i]
								+ "|")) {

							// Check that a value for attribute exists
							if (getRequest().getParameter(parameterName) == null
									|| getRequest().getParameter(parameterName)
											.length() == 0) {
								if (log.isDebugEnabled())
									log.debug("missing required attribute: "
											+ getRequiredAttributeCodes()[j]);
								getErrors().add(
										ActionMessages.GLOBAL_MESSAGE,
										new ActionMessage(
												"errors.requiredAttribute",
												getRequiredAttributeNames()[j],
												getProductNames()[j]));
								getRequest().setAttribute("errorProductCode",
										getOrderProductCodes()[i]);
							}
						}
					}
				}
			}
		}

		// Populating attribute codes and values
		if (getOrderProductCodes() != null) {
			if (log.isDebugEnabled())
				log.debug("Populating attribute codes and values.");
			for (Enumeration e = getRequest().getParameterNames(); e
					.hasMoreElements();) {
				String parameterName = (String) e.nextElement();
				if (log.isDebugEnabled())
					log.debug("Parameter name: " + parameterName);
				for (int i = 0; i < getOrderProductCodes().length; i++) {
					if (log.isDebugEnabled())
						log.debug("getOrderProductCodes()[i]: "
								+ getOrderProductCodes()[i]);
					if (parameterName.startsWith(getOrderProductCodes()[i]
							+ "|")) {
						String[] parameterValues = getRequest()
								.getParameterValues(parameterName);

						// Enforce the attribute max length setting
						String attributeValueMaxLength = (String) getSettings()
								.getValue("attributeValueMaxLength");
						if (attributeValueMaxLength == null) {
							attributeValueMaxLength = "250";
						}
						int maxLength = Integer
								.parseInt(attributeValueMaxLength);
						for (int j = 0; j < parameterValues.length; j++) {
							if (parameterValues[j].length() > maxLength) {
								parameterValues[j] = parameterValues[j]
										.substring(0, maxLength);
							}
						}
						getOrderAttributes()
								.put(parameterName, parameterValues);
					}
				}
			}
		}

	}

	public void validateMaxItemsPerOrder() {

		// If the setting doesn't exist, skip it
		String maxItemsPerOrder = (String) getSettings().getValue(
				"maxItemsPerOrder");
		if (maxItemsPerOrder == null) {
			return;
		}

		// Get the current number of items in the cart
		int totalInCart = 0;
		if (getUser().getOrder() != null
				&& getUser().getOrder().firstDelivery() != null
				&& getUser().getOrder().firstDelivery()
						.getOrderItemCollection() != null) {
			totalInCart = getUser().getOrder().firstDelivery()
					.getOrderItemCollection().size();
		}

		// Find out how many are being added
		int numberBeingAdded = 0;
		if (getOrderProductCodes() != null
				&& getOrderProductQuantities() != null) {
			for (int i = 0; i < getOrderProductCodes().length; i++) {
				if (getOrderProductQuantities()[i] != null) {
					String[] quantity = { getOrderProductQuantities()[i] };
					for (int j = 0; j < quantity.length; j++) {
						try {
							if (quantity[j] != null
									&& Integer.parseInt(quantity[j]) > 0) {
								numberBeingAdded++;
							}
						} catch (NumberFormatException nfe) {
							if (log.isDebugEnabled())
								log.debug(nfe.getMessage());
						}
					}
				}
			}
		}

		// Compare with the max items allowed in the cart
		int maxAllowed = Integer.parseInt(maxItemsPerOrder);
		if (numberBeingAdded > 0 && totalInCart + numberBeingAdded > maxAllowed) {

			// Add the error message
			getErrors().add(
					ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("errors.tooManyItemsInCart",
							maxItemsPerOrder));
			getRequest().setAttribute("errorProductCode", "");
			if (getOrderProductCodes() != null
					&& getOrderProductCodes().length > 0) {
				getRequest().setAttribute("errorProductCode",
						getOrderProductCodes()[0]);
			}
		}

	}
}
